###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""TransformationPlugin is a class wrapping the supported LHCb transformation
plugins."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=too-many-lines
# pylint: disable=missing-docstring

__RCSID__ = "$Id$"

import time
import random
import sys
import six
from collections import defaultdict

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import breakListIntoChunks, randomize
# from DIRAC.Core.Utilities.Time import timeThis
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient
from DIRAC.TransformationSystem.Agent.TransformationPlugin import TransformationPlugin as DIRACTransformationPlugin
from DIRAC.TransformationSystem.Client.Utilities import getFileGroups, sortExistingSEs

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import PluginUtilities, getActiveSEs


class TransformationPlugin(DIRACTransformationPlugin):
  """ Extension of DIRAC TransformationPlugin - instantiated by the TransformationAgent
  """

  def __init__(self, plugin,
               transClient=None, dataManager=None,
               bkClient=None, rmClient=None, fc=None,
               debug=False, transInThread=None):
    """The clients can be passed in."""
    super(TransformationPlugin, self).__init__(plugin,
                                               transClient=transClient,
                                               dataManager=dataManager,
                                               fc=fc)

    if not bkClient:
      self.bkClient = BookkeepingClient()
    else:
      self.bkClient = bkClient

    if not rmClient:
      self.rmClient = ResourceManagementClient()
    else:
      self.rmClient = rmClient

    if not fc:
      self.fileCatalog = FileCatalog()
    else:
      self.fileCatalog = fc

    self.params = {}
    self.workDirectory = None
    self.pluginCallback = self.voidMethod
    self.startTime = time.time()
    self.transReplicas = {}
    self.transFiles = []
    self.transID = None
    self.debug = debug
    if transClient is None:
      self.transClient = TransformationClient()
    else:
      self.transClient = transClient
    self.util = PluginUtilities(plugin=plugin,
                                transClient=transClient, dataManager=dataManager,
                                bkClient=self.bkClient, rmClient=self.rmClient,
                                debug=debug, transInThread=transInThread if transInThread else {})
    self.setDebug(self.util.getPluginParam('Debug', False))

    self.processingShares = ({}, {})
    self._alreadyProcessedLFNs = {}

  def voidMethod(self, _id, invalidateCache=False):
    return

  def setInputData(self, data):
    """self.transReplicas are the replica location of the transformation files.

    However if some don't have a replica, they are not in this
    dictionary self.transReplicas[lfn] == [ SE1, SE2...]
    """
    # data is a synonym as used in DIRAC
    self.transReplicas = data.copy()
    self.data = self.transReplicas
    self.util.transReplicas = self.transReplicas

  def setTransformationFiles(self, files):
    """
    self.transFiles are all the Unused files for that transformation
    It is a list of dictionaries, of which ``lfn = fileDict['LFN']``

    Keys are: ``ErrorCount``, ``FileID``, ``InsertedTime``, ``LFN``, ``LastUpdate``,
      ``RunNumber``, ``Status``, ``TargetSE``, ``TaskID``, ``TransformationID`` and ``UsedSE``
    """
    # files is a synonym, as used in DIRAC
    self.transFiles = [fileDict for fileDict in files]
    self.files = self.transFiles
    self.util.transFiles = self.transFiles

  def setParameters(self, params):
    super(TransformationPlugin, self).setParameters(params)
    self.transID = params['TransformationID']
    self.setDebug(self.util.getPluginParam('Debug', False))

  def setDebug(self, val=True):
    self.debug = val or self.debug
    self.util.setDebug(val)

  def __del__(self):
    self.util.logInfo("Execution finished, timing: %.3f seconds" % (time.time() - self.startTime))

  # @timeThis
  def _removeProcessedFiles(self):
    """Checks if the LFNs have descendants in the same transformation.

    Removes them from self.transReplicas and sets them 'Processed'
    """
    self.util.logVerbose('Checking if %d files are processed' % len(self.transReplicas))
    descendants = self.util.getProcessedFiles(list(self.transReplicas))
    if descendants:
      processedLfns = [lfn for lfn in descendants if descendants[lfn]]
      self.util.logVerbose("Found %d input files that have already been processed (setting status)" %
                           len(processedLfns))
      if processedLfns:
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Processed', processedLfns)
        if res['OK']:
          self.util.logInfo("Found %d input files that have already been processed (status set)" % len(processedLfns))
          for lfn in processedLfns:
            self.transReplicas.pop(lfn, None)
          for fileDict in [fileDict for fileDict in self.transFiles if fileDict['LFN'] in processedLfns]:
            self.transFiles.remove(fileDict)
      else:
        # Here one should check descendants of children
        self.util.logVerbose("No input files have already been processed")

  def _RAWReplication(self):
    """Plugin for replicating RAW data to Tier1s according to shares, and
    defining the processing destination site."""
    self.util.logInfo("Starting execution of plugin")
    sourceSE = 'CERN-RAW'
    rawTargets = self.util.getPluginParam('RAWStorageElements', ['Tier1-RAW'])
    rawTargets = set(resolveSEGroup(rawTargets)) - set([sourceSE])
    bufferTargets = self.util.getPluginParam('ProcessingStorageElements', ['Tier1-Buffer'])
    bufferTargets = set(resolveSEGroup(bufferTargets))
    useRunDestination = self.util.getPluginParam('UseRunDestination', False)
    preStageShares = self.util.getPluginParam('PrestageShares', 'CPUforRAW')
    if preStageShares not in ('CPUforRAW', 'CPUforReprocessing') or not bufferTargets:
      self.util.logInfo('No prestaging required')
      preStageShares = None
    self.util.logVerbose('Targets for replication are %s and %s' % (sorted(rawTargets), sorted(bufferTargets)))
    if preStageShares:
      self.util.logInfo("Using prestage shares from %s" % preStageShares)

    # Get the requested shares from the CS
    res = self.util.getPluginShares(section='RAW')
    if not res['OK']:
      self.util.logError("Section RAW in Shares not available")
      return res
    existingCount, targetShares = res['Value']

    runFileDict = self.util.getFilesGroupedByRun()
    if not runFileDict:
      return S_OK([])

    # For each of the runs determine the destination of any previous files
    res = self.util.getTransformationRuns(runFileDict)
    if not res['OK']:
      self.util.logError("Error when getting transformation runs",
                         "for runs %s : %s" % (','.join(sorted(runFileDict)), res['Message']))
      return res
    runSEDict = dict((runDict['RunNumber'], runDict['SelectedSite']) for runDict in res['Value'])

    # Choose the destination SE
    tasks = []
    alreadyReplicated = []
    for runID in set(runFileDict) & set(runSEDict):
      bufferLogged = False
      rawLogged = False
      # It may happen that files are not in the replica dictionary
      runLfns = set(runFileDict[runID]) & set(self.transReplicas)
      if not runLfns:
        continue
      assignedSE = runSEDict[runID]
      if assignedSE:
        # We already know where this run should go
        ses = set(assignedSE.split(','))
        assignedRAW = ses & rawTargets
        assignedBuffer = ses & bufferTargets
      else:
        assignedRAW = None
        assignedBuffer = None
      # Now determine where these files should go
      # Group by location
      replicaGroups = getFileGroups(dict((lfn, self.transReplicas[lfn])
                                         for lfn in runLfns if lfn in self.transReplicas))
      runSEs = set()
      for replicaSE in replicaGroups:
        # Get all locations where files are
        ses = set(replicaSE.split(','))
        runSEs.update(ses)
      # Find potential locations
      if not assignedRAW:
        assignedRAW = runSEs & rawTargets
      if assignedRAW:
        assignedRAW = list(assignedRAW)[0]
      if not assignedBuffer and preStageShares:
        assignedBuffer = runSEs & bufferTargets
      if assignedBuffer:
        assignedBuffer = list(assignedBuffer)[0]
      updated = False
      for replicaSE, lfns in replicaGroups.items():  # can be an iterator
        replicaSE = set(replicaSE.split(','))
        if not assignedRAW:
          # Files are not yet at a Tier1-RAW, if a destination already exists, use it
          assignedRAW = self.util.getSEForDestination(runID, rawTargets)
          if assignedRAW:
            self.util.logVerbose('RAW destination obtained from run %d destination: %s' % (runID, assignedRAW))
          elif not useRunDestination:
            # Define the destination only if not requested to use it from the run destination
            res = self._getNextSite(existingCount, targetShares)
            if not res['OK']:
              self.util.logError("Failed to get next destination SE", res['Message'])
              return res
            else:
              assignedRAW = res['Value']
              self.util.logVerbose("RAW destination assigned for run %d: %s" % (runID, assignedRAW))
          elif rawTargets:
            self.util.logVerbose("Run destination not yet defined for run %d" % runID)
            # We can go to next in the loop there
            continue
          rawLogged = True
        elif not rawLogged:
          self.util.logVerbose('RAW destination existing for run %d: %s' % (runID, assignedRAW))

        # # Now get a buffer destination is prestaging is required
        if preStageShares and not assignedBuffer:
          assignedBuffer = self.util.getSEForDestination(runID, bufferTargets)
          if assignedBuffer:
            bufferLogged = True
            self.util.logVerbose('Buffer destination obtained from run %d destination: %s' % (runID, assignedBuffer))
          elif not useRunDestination:
            # Files are not at a buffer for processing
            res = self._selectRunSite(runID, sourceSE, replicaSE | set([assignedRAW]),
                                      bufferTargets, preStageShares=preStageShares)
            if not res['OK']:
              self.util.logError("Error selecting the destination site", res['Message'])
              return res
            assignedBuffer = res['Value']
            if assignedBuffer:
              bufferLogged = True
              self.util.logVerbose('Selected destination SE for run %d: %s' % (runID, assignedBuffer))
            else:
              self.util.logWarn('Failed to find Buffer destination SE for run', str(runID))
              continue
          elif bufferTargets:
            self.util.logVerbose("Run destination not yet defined for run %d" % runID)
        elif assignedBuffer and not bufferLogged:
          self.util.logVerbose('Buffer destination existing for run %d: %s' % (runID, assignedBuffer))

        # # Find out if the replication is necessary
        assignedSE = []
        if assignedRAW:
          assignedSE.append(assignedRAW)
        if assignedBuffer:
          assignedSE.append(assignedBuffer)
        if assignedSE:
          if not updated:
            updated = True
            res = self.transClient.setTransformationRunsSite(self.transID, runID, ','.join(assignedSE))
            if not res['OK']:
              self.util.logError("Failed to assign TransformationRun SE", res['Message'])
              return res
          ses = sorted(set(assignedSE) - replicaSE)
          # Update the counters as we know the number of files
          if assignedRAW in ses:
            # Here we pass both the number of files and the runID as we can use either metrics
            self.util.updateSharesUsage(existingCount, assignedRAW, len(lfns), runID)
          assignedSE = ','.join(ses)
        if assignedSE:
          self.util.logVerbose('Creating a task (%d files, run %d) for SEs %s' %
                               (len(lfns), runID, assignedSE))
          tasks.append((assignedSE, lfns))
        else:
          alreadyReplicated += lfns
          self.util.logVerbose('%d files in run %d found already replicated at %s' % (len(lfns),
                                                                                      runID,
                                                                                      ','.join(sorted(replicaSE))))

    if alreadyReplicated:
      for lfn in alreadyReplicated:
        self.transReplicas.pop(lfn)
      self.util.cleanFiles(self.transFiles, self.transReplicas, status='Processed')

    self.util.printShares("Final target shares and usage (%)",
                          targetShares,
                          existingCount,
                          log=self.util.logVerbose)
    return S_OK(tasks)

  def _RAWProcessing(self):
    """Create tasks for RAW data processing using the run destination table."""
    # Let's create jobs only at active SEs
    fromSEs = set(se for se in resolveSEGroup(self.util.getPluginParam('FromSEs', []))
                  if StorageElement(se).status()['Read'])
    if not fromSEs:
      self.util.logWarn('No processing active SEs are provided')
      return S_OK([])

    # Split the files in run groups
    runFileDict = self.util.getFilesGroupedByRun()
    if not runFileDict:
      # No files, no tasks!
      self.util.logVerbose("No runs found!")
      return S_OK([])
    self.util.logVerbose('Obtained %d runs' % len(runFileDict))

    # For each of the runs determine the destination of any previous files
    res = self.util.getTransformationRuns(runFileDict)
    if not res['OK']:
      self.util.logError("Error when getting transformation runs",
                         "for runs %s : %s" % (','.join(sorted(runFileDict)), res['Message']))
      return res
    runSEDict = dict((runDict['RunNumber'], runDict['SelectedSite']) for runDict in res['Value'])

    # Get status of all runs (finished or not)
    runSet = set(runFileDict) & set(runSEDict)
    res = self.bkClient.getRunStatus(list(runSet))
    success = res.get('Value', {}).get('Successful', {})
    runFinished = dict((runID, success[runID]['Finished'] == 'Y') for runID in success)

    # Choose the destination SE
    tasks = []
    for runID in runSet:
      assignedSE = runSEDict[runID]
      runSEs = set(assignedSE.split(',')) if assignedSE and isinstance(assignedSE, six.string_types) else set()
      if not runSEs:
        # Check that the run tick is present, but if an SE is already assigned, this is not needed
        runTick = self.util.checkCondDBRunTick(runID)
        if runTick is False:
          # If runTick is None it means the CondDB was not available in CVMFS, we ignore (but a warning is printed)
          self.util.logInfo("Run tick not present yet in OnlineCondDB for run %d" % runID)
          continue
      runLfns = runFileDict[runID]
      # Now determine where these files should go
      # Group by location
      update = False
      replicaGroups = getFileGroups(dict((lfn, self.transReplicas[lfn])
                                         for lfn in runLfns if lfn in self.transReplicas))
      notAtSE = 0
      for replicaSE, lfns in replicaGroups.items():  # can be an iterator
        targetSEs = set(replicaSE.split(',')) & fromSEs
        if targetSEs:
          # The files are at at least one of the requested SEs, set in run site for transformation
          #   and create task
          update = targetSEs - runSEs
          if update:
            runSEs |= targetSEs
          targetSEs = ','.join(sorted(targetSEs))
          self.util.logVerbose('Creating tasks with %d files for run %s at %s' % (len(lfns), runID, targetSEs))
          # We flush the run only if it is finished
          toFlush = (self.params['Status'] == 'Flush') or runFinished.get(runID, False)
          newTasks = self.util.createTasksBySize(lfns, targetSEs, flush=toFlush)
          tasks += newTasks
          self.util.logVerbose('Created %d tasks' % len(newTasks))
        else:
          notAtSE += len(lfns)
      if notAtSE:
        self.util.logVerbose('Found %d files not yet at required SEs for run %d' % (notAtSE, runID))
      # If there are new run site destination SEs, set them
      if update:
        res = self.transClient.setTransformationRunsSite(self.transID, runID, ','.join(sorted(runSEs)))
        if not res['OK']:
          self.util.logError("Failed to assign TransformationRun SE", res['Message'])
          return res

    if self.pluginCallback:
      self.pluginCallback(self.transID, invalidateCache=True)
    return S_OK(tasks)

  def _selectRunSite(self, runID, backupSE, rawSEs, bufferTargets, preStageShares=None):
    if not preStageShares:
      return S_OK()

    if not self.processingShares[0]:
      res = self.util.getPluginShares(section=preStageShares, backupSE=backupSE)
      if not res['OK']:
        self.util.logError("Error getting CPU shares for RAW processing", res['Message'])
        return res
      else:
        self.processingShares = res['Value']

    self.util.logVerbose("Select processing SE for run %d within %s" % (runID, sorted(rawSEs)))
    rawFraction, cpuShares = self.processingShares
    if len(rawSEs) == 1:
      selectedSE = list(rawSEs)[0]
    else:
      existingSEs = set(cpuShares) & rawSEs
      if not existingSEs:
        errStr = "Could not find shares for SEs"
        self.util.logError(errStr, sorted(rawSEs))
        return S_ERROR(errStr)

      prob = 0
      seProbs = {}
      rawSEs = sorted((rawSEs & set(rawFraction)) - set([backupSE]))
      for se in rawSEs:
        prob += rawFraction[se] / len(rawSEs)
        seProbs[se] = prob
      rawSEs.append(backupSE)
      seProbs[backupSE] = 1.
      rand = random.uniform(0., 1.)
      strProbs = ','.join([' %s:%.3f' % (se, seProbs[se]) for se in rawSEs])
      self.util.logInfo("For run %d, SE integrated fraction =%s, random number = %.3f" % (runID, strProbs, rand))
      selectedSE = None
      for se in rawSEs:
        prob = seProbs[se]
        if rand <= prob:
          selectedSE = se
          break

    # Find out the site associated to that SE
    res = self.util.dmsHelper.getLocalSiteForSE(selectedSE)
    if not res['OK']:
      return res
    site = res['Value']
    if site is None:
      return S_ERROR("No site found for SE")
    res = self.transClient.setDestinationForRun(runID, site)
    if not res['OK']:
      return res
    self.util.logVerbose('Successfully set destination for run %d: %s' % (runID, site))

    return self.util.dmsHelper.getSEInGroupAtSite(bufferTargets, site)

  def _groupBySize(self, files=None):
    """Generate a task for a given amount of data at a (set of) SE."""
    if not files:
      files = self.transReplicas
    else:
      files = dict(zip(files, [self.transReplicas[lfn] for lfn in files]))
    return self.util.groupBySize(files, self.params['Status'])

  def _LHCbStandard(self):
    """Plugin grouping files at same sites based on number of files, used for
    example for stripping or WG productions."""
    return self.util.groupByReplicas(self.transReplicas, self.params['Status'])

  def _ByRun(self, param='', plugin='LHCbStandard', requireFlush=False, forceFlush=False):
    try:
      return self.__byRun(param=param, plugin=plugin, requireFlush=requireFlush, forceFlush=forceFlush)
    except Exception as x:
      self.util.logException("Exception in _ByRun plugin:", '', lException=x)
      return S_ERROR([])

  # @timeThis
  def __byRun(self, param='', plugin='LHCbStandard', requireFlush=False, forceFlush=False):
    """Basic plugin for when you want to group files by run."""
    self.util.logInfo("Starting execution of plugin")
    allTasks = []
    if not self.transReplicas:
      self.util.logVerbose("No data to be processed by plugin")
      return S_OK(allTasks)
    # If flush is force, it is obviously required! try and get forcing from parameter
    if self.util.getPluginParam('ForceFlush', False):
      forceFlush = True
    requireFlush |= forceFlush
    pluginStartTime = time.time()
    groupSize = self.util.getPluginParam('GroupSize')
    typesWithNoCheck = self.util.getPluginParam('NoCheckTypes',
                                                ['Merge', 'MCMerge', 'HistoMerge', 'Replication', 'Removal'])
    # Only consider active SEs for read
    fromSEs = set(se for se in resolveSEGroup(self.util.getPluginParam('FromSEs', []))
                  if StorageElement(se).status()['Read'])
    maxTime = self.util.getPluginParam('MaxTimeAllowed', 0)

    # Check if ancestors are required: this flag defaults to True for DataStripping transformations
    lfn = list(self.transReplicas)[0]
    fileType = self.util.getMetadataFromTSorBK(lfn, 'FileType')[lfn]
    addAncestors = self.util.getPluginParam('UseAncestors',
                                            bool(self.params['Type'] == 'DataStripping') and fileType != 'FULL.DST')
    # Read the cached information from disk
    self.util.readCacheFile(self.workDirectory)

    # Group files by run number and value of parameter "param"
    self.util.logInfo("Grouping %d files by runs %s " %
                      (len(self.transFiles), 'and %s' % param if param else ''))
    runFileDict = self.util.getFilesGroupedByRunAndParam(param=param)

    transStatus = self.params['Status']
    res = self.util.getTransformationRuns(runFileDict)
    if not res['OK']:
      self.util.logError("Error when getting transformation runs",
                         "for runs %s : %s" % (','.join(sorted(runFileDict)), res['Message']))
      return res
    transRuns = res['Value']
    runSites = dict((run['RunNumber'], run['SelectedSite']) for run in transRuns if run['SelectedSite'])
    # Loop on all runs that have new files
    inputData = self.transReplicas.copy()
    setInputData = set(inputData)
    runEvtType = {}
    # Restart where we finished last time and wrap to beginning
    lastRun = self.util.getCachedLastRun()
    runNumbers = sorted(run['RunNumber'] for run in transRuns if run['RunNumber'] > lastRun) + \
        sorted(run['RunNumber'] for run in transRuns if run['RunNumber'] <= lastRun)
    # Find out how many files we have currently Unused per run
    res = self.transClient.getTransformationFilesCount(self.transID, 'RunNumber',
                                                       {'Status': 'Unused', 'RunNumber': runNumbers})
    if not res['OK']:
      self.util.logError("Error getting file counts per run", res['Message'])
      return res
    unusedFilesPerRun = res['Value']
    # Check that the total number of files we got for that run is equal to the number of Unused files
    nRunsLeft = len(runNumbers)
    for runID in list(runNumbers):
      runFiles = sum(len(lfns) for lfns in runFileDict[runID].values())  # can be an iterator
      if runFiles != unusedFilesPerRun.get(runID, 0):
        runNumbers.remove(runID)
    if nRunsLeft != len(runNumbers):
      self.util.logWarn("Removed %d runs with less files than Unused" % (nRunsLeft - len(runNumbers)))
      nRunsLeft = len(runNumbers)
    # Get the list of runs in the same order as runNumbers
    runList = sorted((run for run in transRuns if run['RunNumber'] in runNumbers),
                     cmp=(lambda d1, d2: runNumbers.index(d1['RunNumber']) - runNumbers.index(d2['RunNumber'])))
    if nRunsLeft:
      self.util.logInfo("Processing %d runs between runs %d and %d, starting at run %d" % (nRunsLeft,
                                                                                           min(runNumbers),
                                                                                           max(runNumbers),
                                                                                           runNumbers[0]))
    else:
      self.util.logInfo("No runs to process, exit")
      return S_OK([])
    #
    # # # # # # # Loop on all selected runs # # # # # # #
    #
    timeout = False
    processedFiles = 0
    missingAtSEs = False
    for run in runList:
      runID = run['RunNumber']
      self.util.logVerbose("Processing run %d, still %d runs left" % (runID, nRunsLeft))
      nRunsLeft -= 1
      runStatus = 'Flush' if transStatus == 'Flush' else run['Status']
      paramDict = runFileDict.get(runID, {})
      runTargetSEs = set(se for se in runSites.get(runID, '').split(',') if se)
      #
      # Loop on parameters (None if not by param)
      #
      flushed = []
      for paramValue in sorted(paramDict):
        paramStr = " (%s : %s) " % (param, paramValue) if paramValue else ' '
        runParamLfns = set(paramDict[paramValue])
        processedFiles += len(runParamLfns)
        # Check if something was new since last time...
        cachedLfns = self.util.getCachedRunLFNs(runID, paramValue)
        newLfns = runParamLfns - cachedLfns
        checkForNewFiles = not forceFlush and \
            self.transID > 0 and \
            runStatus != 'Flush' and \
            not self.util.cacheExpired(runID) and \
            (plugin != 'LHCbStandard' or groupSize != 1)

        if checkForNewFiles and \
           not len(newLfns):
          self.util.logInfo("No new files since last time for run %d%s: skip..." % (runID, paramStr))
          continue
        self.util.logVerbose("Of %d files, %d are new for %d%s" % (len(runParamLfns),
                                                                   len(newLfns), runID, paramStr))
        runFlush = requireFlush
        if runFlush:
          # Determine which event type is used, in order to get the corresponding number of RAW files
          if not runEvtType.get(paramValue):
            lfn = paramDict[paramValue][0]
            runEvtType[paramValue] = self.util.getMetadataFromTSorBK(lfn, 'EventType')[lfn]
            self.util.logDebug('Event type%s: %s' % (paramStr, str(runEvtType[paramValue])))
          evtType = runEvtType[paramValue]
          if not evtType:
            runFlush = False
        runParamReplicas = {}
        notAtSEs = 0
        for lfn in runParamLfns & setInputData:
          runParamReplicas[lfn] = [se for se in inputData[lfn]
                                   if self.util.dmsHelper.isSEForJobs(se, checkSE=False) and
                                   (not fromSEs or se in fromSEs)]
          if not runParamReplicas[lfn]:
            notAtSEs += 1
            del runParamReplicas[lfn]
        if notAtSEs:
          missingAtSEs = True
          self.util.logInfo("For run %d, %d files are not at required SEs: tasks cannot be created" % (runID,
                                                                                                       notAtSEs))

        # We need to replace the input replicas by those of this run before calling the helper plugin
        # As it may use self.data, set both transReplicas and data members
        self.transReplicas = runParamReplicas
        # Check if files have already been processed
        # Skipped as this is taking quite some time... Leave it here as reminder, in case it becomes mandatory
        if False and self.params['Type'] not in typesWithNoCheck:
          self.util.logInfo("Removing processed files for %s" % paramStr)
          self._removeProcessedFiles()
        # Backward compatibility for calling helper plugin
        self.data = self.transReplicas
        status = runStatus
        if status != 'Flush' and runFlush:
          # If all files in that run have been processed and received, flush
          # Get the number of RAW files in that run
          if not forceFlush:
            retried = False
            ancestorRawFiles = self.util.getRAWAncestorsForRun(runID, param, paramValue)
            self.util.logVerbose("Obtained %d ancestor RAW files" % ancestorRawFiles)
            while True:
              rawFiles = self.util.getNbRAWInRun(runID, evtType)
              if not retried and rawFiles and ancestorRawFiles > rawFiles:
                # In case there are more ancestors than RAW files
                # we may have to refresh the number of RAW files: try once
                self.util.cachedNbRAWFiles[runID][evtType] = 0
                retried = True
              else:
                runProcessed = (ancestorRawFiles == rawFiles)
                break
          else:
            runProcessed = False
          if forceFlush or runProcessed:
            if runProcessed:
              # The whole run was processed by the parent production and we received all files
              self.util.logInfo("All RAW files (%d) ready for run %d%s- Flushing run" %
                                (rawFiles, runID, paramStr))
            status = 'Flush'
            runStatus = status
            self.transClient.setTransformationRunStatus(self.transID, runID, 'Flush')
          elif rawFiles:
            self.util.logVerbose("Only %d ancestor RAW files (of %d) available for run %d" %
                                 (ancestorRawFiles, rawFiles, runID))
        if runStatus == 'Flush':
          flushed.append((paramValue, len(self.transReplicas)))
        # Now calling the helper plugin... Set status to a fake value
        self.params['Status'] = status
        res = eval('self._%s()' % plugin)  # pylint: disable=eval-used
        # Resetting status
        self.params['Status'] = transStatus
        if not res['OK']:
          return res
        tasks = res['Value']
        if fromSEs:
          # The files and if needed its ancestors have to be in a set of SEs
          # This is used to force processing from disk replicas
          missingAncestors = 0
          okDict = {}
          # Group the LFNs by (set of) SEs in order to speed up the check for ancestors
          nbTasks = len(tasks)
          for task in tasks:
            # Restrict target SEs to those in fromSEs
            okSEs = fromSEs.intersection(task[0].split(','))
            if okSEs:
              okDict.setdefault(tuple(okSEs), []).append(task[1])
          # Create the real tasks now
          tasks = []
          for okSEs, taskLfns in okDict.items():  # can be an iterator
            if addAncestors:
              # taskLfns is modified by this method: lfns are eventually removed
              missing = self.util.checkAncestorsAtSE(taskLfns, fromSEs)
              if missing:
                missingAncestors += missing
            # Now create the list of tasks
            tasks += [(','.join(sorted(okSEs)), lfnList) for lfnList in taskLfns if lfnList]
          # Log if files are not at the required SEs
          if missingAncestors:
            missingAtSEs = True
            self.util.logInfo("%d files have been removed from tasks as ancestors were not present at required SEs" %
                              missingAncestors)
          if nbTasks != len(tasks):
            missingAtSEs = True
            self.util.logInfo("%d tasks could not be created for run %d as files are not at required SEs" %
                              (nbTasks - len(tasks), runID))
        self.util.logInfo("Created %d tasks for run %d%s" %
                          (len(tasks), runID, paramStr))
        allTasks.extend(tasks)
        taskLfns = set(lfn for task in tasks for lfn in task[1])
        # If there are new SEs used for that run, add them to the DB
        targetSEs = set(se for task in tasks for se in task[0].split(','))
        if targetSEs - runTargetSEs:
          targetSEs = ','.join(targetSEs | runTargetSEs)
          self.util.logVerbose("Set target SEs for run %d as %s" % (runID, targetSEs))
          res = self.transClient.setTransformationRunsSite(self.transID, runID, targetSEs)
          if not res['OK']:
            self.util.logError("Failed to set target SEs", "to run %d as %s: %s" %
                               (runID, targetSEs, res['Message']))
        # Cache the remaining files
        self.util.setCachedRunLfns(runID, paramValue, set(runParamLfns) - taskLfns)
      # # # # # # # # # # #
      # End of param loop #
      # # # # # # # # # # #
      if flushed:
        # Print out info about what was flushed
        prStr = "Run %d is flushed%s:" % (runID, " for %s" % param if param else "")
        prStr += ','.join(" %s (%d files)" % flTuple
                          if flTuple[0]
                          else " %d files" % flTuple[1]
                          for flTuple in sorted(flushed))
        self.util.logInfo(prStr)

      # if enough time already spent, exit
      timeSpent = time.time() - pluginStartTime
      lastRun = runID
      if maxTime and timeSpent > maxTime:
        timeout = True
        self.util.logInfo("Enough time spent in plugin (%.1f seconds), exit at run %d, %d runs left" %
                          (timeSpent, runID, nRunsLeft))
        break
    # # # # # # # # # #
    # End of run loop #
    # # # # # # # # # #
    timeSpent = time.time() - pluginStartTime
    self.util.logInfo("Processed %d files in %.1f seconds" % (processedFiles, timeSpent))
    self.util.setCachedLastRun(lastRun)
    # reset the input data as it was when calling the plugin
    self.setInputData(inputData)
    self.util.writeCacheFile()
    if missingAtSEs and self.pluginCallback:
      # If some files could not be scheduled, clear the cache
      self.pluginCallback(self.transID, invalidateCache=True)
    ret = S_OK(allTasks)
    ret['Timeout'] = timeout
    return ret

  def _ByRunWithFlush(self):
    # If groupSize is 1, no need to flush!
    groupSize = self.util.getPluginParam('GroupSize')
    return self._ByRun(requireFlush=groupSize != 1)

  def _ByRunForceFlush(self):
    # If groupSize is 1, no need to flush!
    groupSize = self.util.getPluginParam('GroupSize')
    return self._ByRun(forceFlush=groupSize != 1)

  def _ByRunSize(self):
    return self._ByRun(plugin='BySize')

  def _ByRunSizeWithFlush(self):
    return self._ByRun(plugin='BySize', requireFlush=True)

  def _ByRunSizeForceFlush(self):
    return self._ByRun(plugin='BySize', forceFlush=True)

  def _ByRunFileType(self):
    return self._ByRun(param='FileType')

  def _ByRunFileTypeWithFlush(self):
    # If groupSize is 1, no need to flush!
    groupSize = self.util.getPluginParam('GroupSize')
    return self._ByRun(param='FileType', requireFlush=groupSize != 1)

  def _ByRunFileTypeForceFlush(self):
    # If groupSize is 1, no need to flush!
    groupSize = self.util.getPluginParam('GroupSize')
    return self._ByRun(param='FileType', forceFlush=groupSize != 1)

  def _ByRunFileTypeSize(self):
    return self._ByRun(param='FileType', plugin='BySize')

  def _ByRunFileTypeSizeWithFlush(self):
    return self._ByRun(param='FileType', plugin='BySize', requireFlush=True)

  def _ByRunFileTypeSizeForceFlush(self):
    return self._ByRun(param='FileType', plugin='BySize', forceFlush=True)

  def _ByRunEventType(self):
    return self._ByRun(param='EventTypeId')

  def _ByRunEventTypeWithFlush(self):
    # If groupSize is 1, no need to flush!
    groupSize = self.util.getPluginParam('GroupSize')
    return self._ByRun(param='EventTypeId', requireFlush=groupSize != 1)

  def _ByRunEventTypeForceFlush(self):
    # If groupSize is 1, no need to flush!
    groupSize = self.util.getPluginParam('GroupSize')
    return self._ByRun(param='EventTypeId', forceFlush=groupSize != 1)

  def _ByRunEventTypeSize(self):
    return self._ByRun(param='EventTypeId', plugin='BySize')

  def _ByRunEventTypeSizeWithFlush(self):
    return self._ByRun(param='EventTypeId', plugin='BySize', requireFlush=True)

  def _ByRunEventTypeSizeForceFlush(self):
    return self._ByRun(param='EventTypeId', plugin='BySize', forceFlush=True)

  # Plugins for RootMerging -
  # here only to distinguish from the others as "RootMerging" is meaningful at production creation

  def _ByRunSizeWithFlushRootMerging(self):
    return self._ByRun(plugin='BySize', requireFlush=True)

  def _BySizeRootMerging(self):
    return self._BySize()

  def _LHCbDSTBroadcast(self):
    """This plug-in broadcasts files according to CS settings
    to one archiveSE (if set) and <numberOfCopies> secondarySEs
    One can force some mandatorySEs and exclude some SEs
    All files for the same run have the same target
    Usually for replication of real data (2 copies)"""
    archiveSEs = resolveSEGroup(self.util.getPluginParam('ArchiveSEs', []))
    mandatorySEs = resolveSEGroup(self.util.getPluginParam('MandatorySEs', []))
    # In order to not have to change the SEGroups when excluding temporarily a site, add exclusion list...
    excludedSEs = resolveSEGroup(self.util.getPluginParam('ExcludedSEs', []))
    secondarySEs = list(set(resolveSEGroup(self.util.getPluginParam('SecondarySEs', []))) - set(excludedSEs))
    numberOfCopies = self.util.getPluginParam('NumberOfReplicas', 2)

    self.util.logInfo("Starting execution of plugin")

    # We need at least all mandatory copies
    numberOfCopies = max(numberOfCopies, len(mandatorySEs))

    # Group the remaining data by run
    runFileDict = self.util.getFilesGroupedByRun()
    if not runFileDict:
      # No files, no tasks!
      return S_OK([])

    # For each of the runs determine the destination of any previous files
    runSEDict = {}
    runUpdate = {}
    # Make a list of SEs already assigned to runs
    res = self.util.getTransformationRuns(runFileDict)
    if not res['OK']:
      self.util.logError("Error when getting transformation runs",
                         "for runs %s : %s" % (','.join(sorted(runFileDict)), res['Message']))
      return res
    for runDict in res['Value']:
      runID = runDict['RunNumber']
      # If the run already has a selected site, use it for that run
      if runDict['SelectedSite']:
        runSEDict[runID] = runDict['SelectedSite']
        runUpdate[runID] = False
      else:
        # Check if some files are assigned to an SE in this run
        res = self.transClient.getTransformationFiles(condDict={'TransformationID': self.transID,
                                                                'RunNumber': runID,
                                                                'Status': ['Assigned', 'Processed']})
        if not res['OK']:
          self.util.logError("Failed to get transformation files for run", "%s %s" % (runID, res['Message']))
        else:
          if res['Value'] and res['Value'][0]['UsedSE']:
            runSEDict[runID] = res['Value'][0]['UsedSE']
            runUpdate[runID] = True

    fileTargetSEs = {}
    alreadyCompleted = []
    # Consider all runs in turn
    for runID, runLfns in runFileDict.items():  # can be an iterator
      # Check if the run is already assigned
      stringTargetSEs = runSEDict.get(runID, None)
      # No SE assigned yet, determine them
      if not stringTargetSEs:
        # Sort existing SEs where most of the files are already
        existingSEs = sortExistingSEs(self.transReplicas, runLfns)
        # this may happen when all files are in FAILOVER
        if existingSEs:
          # Now select the target SEs
          self.util.logVerbose("Selecting SEs for %d files" % len(runLfns))
          self.util.logDebug("Files: %s" % str(runLfns))
          stringTargetSEs = self.util.setTargetSEs(numberOfCopies, archiveSEs,
                                                   mandatorySEs, secondarySEs, existingSEs, exclusiveSEs=False)
          runUpdate[runID] = True

      # Update the TransformationRuns table with the assigned SEs (don't continue if it fails)
      if stringTargetSEs:
        if runUpdate[runID]:
          res = self.transClient.setTransformationRunsSite(self.transID, runID, stringTargetSEs)
          if not res['OK']:
            self.util.logError("Failed to assign TransformationRun site", res['Message'])
            return S_ERROR("Failed to assign TransformationRun site")

        # Now assign the individual files to their targets
        runFileTargetSEs, runCompleted = self.util.assignTargetToLfns(runLfns, self.transReplicas, stringTargetSEs)
        alreadyCompleted += runCompleted
        fileTargetSEs.update(runFileTargetSEs)

    # Update the status of the already done files
    if alreadyCompleted:
      self.util.logInfo("Found %s files that are already completed" % len(alreadyCompleted))
      self.transClient.setFileStatusForTransformation(self.transID, 'Processed', alreadyCompleted)

    # Now group all of the files by their target SEs
    storageElementGroups = {}
    for lfn, stringTargetSEs in fileTargetSEs.items():  # can be an iterator
      storageElementGroups.setdefault(stringTargetSEs, []).append(lfn)

    return S_OK(self.util.createTasks(storageElementGroups))

  def _LHCbWGBroadcastRandom(self):
    """ This plugin is specific for randomly broadcasting real data files that don't have a run number
    It calls the MCDSTBroadcast plugin but should have different CS settings
    """
    return self._LHCbMCDSTBroadcastRandom()

  def _LHCbMCDSTBroadcastRandom(self):
    """This plug-in broadcasts files to
    one archiveSE (if set) and to random
    <NumberOfReplicas> secondary SEs."""

    self.util.logInfo("Starting execution of plugin")
    archiveSEs = resolveSEGroup(self.util.getPluginParam('ArchiveSEs', []))
    mandatorySEs = resolveSEGroup(self.util.getPluginParam('MandatorySEs', []))
    # In order to not have to change the SEGroups when excluding temporarily a site, add exclusion list...
    excludedSEs = resolveSEGroup(self.util.getPluginParam('ExcludedSEs', []))
    secondarySEs = list(set(resolveSEGroup(self.util.getPluginParam('SecondarySEs', []))) - set(excludedSEs))
    numberOfCopies = self.util.getPluginParam('NumberOfReplicas', 2)
    excludedFileTypes = self.util.getPluginParam('ExcludedFileTypes', ['GAUSSHIST', 'BRUNELHIST', 'DAVINCIHIST'])

    # We need at least all mandatory copies
    numberOfCopies = max(numberOfCopies, len(mandatorySEs))

    # Filter file types
    if excludedFileTypes:
      excludedLfns = []
      for lfn, fileType in self.util.getMetadataFromTSorBK(
              self.transReplicas, 'FileType').items():  # can be an iterator
        if fileType in excludedFileTypes:
          self.transReplicas.pop(lfn)
          excludedLfns.append(lfn)
      if excludedLfns:
        self.util.logInfo("Found %d files with excluded file type, set them Excluded" % len(excludedLfns))
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Excluded', excludedLfns)
        if not res['OK']:
          self.util.logError("Error setting files Excluded", res['Message'])
          return res

    storageElementGroups = {}
    for replicaSE, lfnGroup in getFileGroups(self.transReplicas).items():  # can be an iterator
      existingSEs = [se for se in replicaSE.split(',') if not self.util.dmsHelper.isSEFailover(se)]
      for lfns in breakListIntoChunks(lfnGroup, 100):

        stringTargetSEs = self.util.setTargetSEs(numberOfCopies, archiveSEs,
                                                 mandatorySEs, secondarySEs, existingSEs, exclusiveSEs=True)
        if stringTargetSEs:
          storageElementGroups.setdefault(stringTargetSEs, []).extend(lfns)
        else:
          self.util.logInfo("Found %d files that are already completed, set them Processed" % len(lfns))
          res = self.transClient.setFileStatusForTransformation(self.transID, 'Processed', lfns)
          if not res['OK']:
            self.util.logError("Error setting files Processed", res['Message'])
            return res

    return S_OK(self.util.createTasks(storageElementGroups))

  def _ReplicateDataset(self, maxFiles=None):
    """Plugin for replicating files to specified SEs."""
    destSEs = resolveSEGroup(self.util.getPluginParam('DestinationSEs', []))
    if not destSEs:
      destSEs = resolveSEGroup(self.util.getPluginParam('MandatorySEs', []))
    secondarySEs = resolveSEGroup(self.util.getPluginParam('SecondarySEs', []))
    fromSEs = resolveSEGroup(self.util.getPluginParam('FromSEs', []))
    numberOfCopies = self.util.getPluginParam('NumberOfReplicas', 0)
    return self._simpleReplication(destSEs, secondarySEs, numberOfCopies, fromSEs=fromSEs, maxFiles=maxFiles)

  def _ReplicateToRunDestination(self):
    """Plugin for replicating files to the run destination."""
    # Get replication throttling parameters and destination
    res = self.util.getMaxFilesToReplicate(self.workDirectory)
    if not res['OK']:
      return res
    watermark, maxFilesAtSE = res['Value']
    # This is a convention, to skip the loop or to stop replicating
    if watermark is None:
      return S_OK([])
    destSEs = list(maxFilesAtSE)
    runFileDict = self.util.getFilesGroupedByRun()

    # Make a list of SEs already assigned to runs
    res = self.util.getTransformationRuns(runFileDict)
    if not res['OK']:
      self.util.logError("Error when getting transformation runs",
                         "for runs %s : %s" % (','.join(sorted(runFileDict)), res['Message']))
      return res
    runSEDict = dict((runDict['RunNumber'], runDict['SelectedSite'])
                     for runDict in res['Value'] if runDict['SelectedSite'])

    maxFiles = self.util.getPluginParam('MaxFilesPerTask', 100)
    tasks = []
    alreadyReplicated = set()
    for runID in runFileDict:
      lfns = set(runFileDict[runID]) & set(self.transReplicas)
      if not lfns:
        continue
      candidateSE = runSEDict.get(runID, self.util.getSEForDestination(runID, destSEs))
      if candidateSE:
        # If necessary, set the run target in the TS
        if runID not in runSEDict:
          res = self.transClient.setTransformationRunsSite(self.transID, runID, candidateSE)
          if not res['OK']:
            self.util.logError("Failed to assign TransformationRun site", res['Message'])
            return S_ERROR("Failed to assign TransformationRun site")
        freeSpace = self.util.getStorageFreeSpace([candidateSE])
        if freeSpace[candidateSE] < watermark:
          self.util.logInfo("No enough space (%s TB) found at %s" % (watermark, candidateSE))
        else:
          replicated = set(lfn for lfn in lfns if candidateSE in self.transReplicas[lfn])
          lfns -= replicated
          alreadyReplicated.update(replicated)
          if lfns:
            maxToReplicate = maxFilesAtSE.get(candidateSE, sys.maxsize)
            if len(lfns) < maxFilesAtSE.get(candidateSE, sys.maxsize):
              self.util.logVerbose("Number of files for %s: %d" % (candidateSE, len(lfns)))
              for lfnChunk in breakListIntoChunks(lfns, maxFiles):
                tasks.append((candidateSE, lfnChunk))
            else:
              self.util.logInfo("Limit number of files for %s to %d (out of %d)" %
                                (candidateSE, maxToReplicate, len(lfns)))

    if alreadyReplicated:
      self.util.logInfo('Found %d files that are already present at destination SE,'
                        ' set them Processed' % len(alreadyReplicated))
      res = self.transClient.setFileStatusForTransformation(self.transID, 'Processed', alreadyReplicated)
      if not res['OK']:
        self.util.logError("Error setting files Processed", res['Message'])
        return res

    return S_OK(tasks)

  def _ArchiveDataset(self):
    """Plugin for archiving datasets, randomly chosen"""
    archiveSEs = resolveSEGroup(self.util.getPluginParam('ArchiveSEs', []))
    numberOfCopies = self.util.getPluginParam('NumberOfReplicas', 1)
    archiveActiveSEs = getActiveSEs(archiveSEs)
    if not archiveActiveSEs:
      archiveActiveSEs = archiveSEs
    return self._simpleReplication([], archiveActiveSEs, numberOfCopies=numberOfCopies)

  def _simpleReplication(self, mandatorySEs, secondarySEs, numberOfCopies=0, fromSEs=None, maxFiles=None):
    """Actually creates the replication tasks for replication plugins."""
    self.util.logInfo("Starting execution of plugin")
    mandatorySEs = set(mandatorySEs)
    secondarySEs = set(secondarySEs) - mandatorySEs
    if not numberOfCopies:
      numberOfCopies = len(secondarySEs) + len(mandatorySEs)
      activeSecondarySEs = secondarySEs
    else:
      activeSecondarySEs = getActiveSEs(secondarySEs)
      numberOfCopies = max(len(mandatorySEs), numberOfCopies)

    self.util.logVerbose("%d replicas, mandatory at %s, optional at %s" % (numberOfCopies,
                                                                           mandatorySEs, activeSecondarySEs))

    alreadyCompleted = []
    fileTargetSEs = {}
    for replicaSE, lfnGroup in getFileGroups(self.transReplicas).items():  # can be an iterator
      existingSEs = [se for se in replicaSE.split(',') if not self.util.dmsHelper.isSEFailover(se)]
      # If a FromSEs parameter is given, only keep the files that are at one of those SEs, mark the others NotProcessed
      if fromSEs:
        if not isinstance(fromSEs, list):
          return S_ERROR("fromSEs parameter should be a list")
        if not set(existingSEs) & set(fromSEs):
          res = self.transClient.setFileStatusForTransformation(self.transID, 'NotProcessed', lfnGroup)
          if not res['OK']:
            self.util.logError('Error setting files NotProcessed', res['Message'])
          else:
            self.util.logVerbose('Found %d files that are not in %s, set NotProcessed' % (len(lfnGroup), fromSEs))
          continue

      # If there is no choice on the SEs, send all files at once, otherwise make chunks
      if numberOfCopies >= len(mandatorySEs) + len(activeSecondarySEs):
        lfnChunks = [lfnGroup]
      else:
        lfnChunks = breakListIntoChunks(lfnGroup, 100)

      self.util.logDebug("Split lfns in %d chunks" % len(lfnChunks))

      for lfns in lfnChunks:
        candidateSEs = self.util.closerSEs(existingSEs, secondarySEs)
        self.util.logDebug("Candidate SEs = %s" % ', '.join(candidateSEs))
        # Remove duplicated SEs (those that are indeed the same), but keep existing ones
        for se1 in [se for se in candidateSEs if se not in existingSEs]:
          if self.util.isSameSEInList(se1, [se for se in candidateSEs if se != se1]):
            candidateSEs.remove(se1)
        # Remove existing SEs from list of candidates
        ncand = len(candidateSEs)
        candidateSEs = [se for se in candidateSEs if se not in existingSEs]
        self.util.logDebug("Candidate SEs after removal = %s" % ', '.join(candidateSEs))
        needToCopy = numberOfCopies - (ncand - len(candidateSEs))
        stillMandatory = [se for se in mandatorySEs if se not in candidateSEs]
        candidateSEs = self.util.uniqueSEs(stillMandatory + [se for se in candidateSEs if se in activeSecondarySEs])
        needToCopy = max(needToCopy, len(stillMandatory))
        targetSEs = []
        if needToCopy > 0:
          if needToCopy <= len(candidateSEs):
            targetSEs = candidateSEs[0:needToCopy]
          else:
            targetSEs = candidateSEs
            needToCopy -= len(targetSEs)
            # Try and replicate to non active SEs
            otherSEs = [se for se in secondarySEs if se not in targetSEs]
            if otherSEs:
              targetSEs += otherSEs[0:min(needToCopy, len(otherSEs))]
        else:
          alreadyCompleted += lfns
        if targetSEs:
          stringTargetSEs = ','.join(sorted(targetSEs))
          # Now assign the individual files to their targets
          chunkFileTargetSEs, completed = self.util.assignTargetToLfns(lfns, self.transReplicas, stringTargetSEs)
          alreadyCompleted += completed
          fileTargetSEs.update(chunkFileTargetSEs)

    # Update the status of the already done files
    if alreadyCompleted:
      self.util.logInfo("Found %s files that are already completed" % len(alreadyCompleted))
      self.transClient.setFileStatusForTransformation(self.transID, 'Processed', alreadyCompleted)

    # Now group all of the files by their target SEs
    storageElementGroups = {}
    for lfn, stringTargetSEs in fileTargetSEs.items():  # can be an iterator
      storageElementGroups.setdefault(stringTargetSEs, []).append(lfn)

    self.util.logDebug("Storage Element Groups created: %s" % storageElementGroups)

    return S_OK(self.util.createTasks(storageElementGroups, chunkSize=maxFiles))

  def _FakeReplication(self):
    """Creates replication tasks for to the existing SEs.

    Used only for tests!
    """
    storageElementGroups = {}
    for replicaSE, lfnGroup in getFileGroups(self.transReplicas).items():  # can be an iterator
      existingSEs = replicaSE.split(',')
      for lfns in breakListIntoChunks(lfnGroup, 100):
        stringTargetSEs = existingSEs[0]
        storageElementGroups.setdefault(stringTargetSEs, []).extend(lfns)
    if self.pluginCallback:
      self.pluginCallback(self.transID, invalidateCache=True)
    return S_OK(self.util.createTasks(storageElementGroups))

  def _DestroyDataset(self):
    """Plugin setting all existing SEs as targets."""
    self.util.logInfo("Starting execution of plugin")
    res = self._removeReplicas(keepSEs=[], minKeep=0)
    if not res['OK'] or not res['Value']:
      return res
    tasks = res['Value']
    if self.util.getPluginParam('CleanTransformations', False):
      lfns = sum((task[1] for task in tasks), [])
      # Check if some of these files are used by transformations
      selectDict = {'LFN': lfns}
      res = self.transClient.getTransformationFiles(selectDict)
      if not res['OK']:
        self.util.logError("Error getting transformation files for %d files" % len(lfns), res['Message'])
      else:
        processedFiles = set()
        self.util.logVerbose("Out of %d files, %d occurrences were found in transformations" % (len(lfns),
                                                                                                len(res['Value'])))
        transDict = defaultdict(list)
        for fileDict in res['Value']:
          # Processed files are immutable, and don't kill yourself!
          if fileDict['TransformationID'] != self.transID:
            if fileDict['Status'] not in ('Processed', 'Removed'):
              transDict[fileDict['TransformationID']].append(fileDict['LFN'])
            else:
              processedFiles.add(fileDict['LFN'])
        if transDict:
          self.util.logVerbose("Files to be set Removed in transformations %s" %
                               ','.join('%d' % trans for trans in sorted(transDict)))
        else:
          self.util.logVerbose("No files to be set Removed in other transformations")
        if processedFiles:
          self.util.logInfo("%d files are being removed but were already"
                            " Processed or Removed in other transformations" %
                            len(processedFiles))
        for trans, lfns in transDict.items():  # can be an iterator
          # Do not actually take action for a fake transformation (dirac-test-plugin)
          if self.transID > 0:
            res = self.transClient.setFileStatusForTransformation(trans, 'Removed', lfns)
            action = 'set'
          else:
            res = {'OK': True, 'Value': lfns}
            action = 'to be set'
          if not res['OK']:
            self.util.logError("Error setting %d files in transformation %d to status Removed" % (len(lfns),
                                                                                                  trans),
                               res['Message'])
          else:
            self.util.logInfo("%d files out of %d %s as 'Removed' in transformation %d" %
                              (len(res['Value']), len(lfns), action, trans))

    return S_OK(tasks)

  def _ReduceReplicasKeepDestination(self):
    """Plugin for reducing the number of replicas to NumberOfReplicas."""
    # this is the number of replicas to be kept in addition to keepSEs and mandatorySEs
    minKeep = -abs(self.util.getPluginParam('NumberOfReplicas', 1))
    return self._RemoveReplicasKeepDestination(minKeep=minKeep)

  def _RemoveReplicasKeepDestination(self, minKeep=None):
    """Plugin used to remove all replicas from a set of SEs except at run
    destination."""
    fromSEs = set(resolveSEGroup(self.util.getPluginParam('FromSEs', [])))
    keepSEs = resolveSEGroup(self.util.getPluginParam('KeepSEs', ['Tier1-Archive']))
    # this is the number of replicas to be kept in addition to keepSEs and mandatorySEs
    # Remove 1 as we keep the run destination as well...
    if minKeep is None:
      minKeep = abs(self.util.getPluginParam('NumberOfReplicas', 1))

    # Group the  data by run
    runFileDict = self.util.getFilesGroupedByRun()
    if not runFileDict:
      # No files, no tasks!
      return S_OK([])

    res = self.util.getTransformationRuns(runFileDict)
    if not res['OK']:
      self.util.logError("Error when getting transformation runs",
                         "for runs %s: %s" % (','.join(str(run) for run in runFileDict), res['Message']))
      return res
    runSites = dict((run['RunNumber'], set(run['SelectedSite'].split(',')))
                    for run in res['Value'] if run['SelectedSite'])

    # Consider all runs in turn
    tasks = []
    for runID, runLfns in runFileDict.items():  # can be an iterator
      replicas = dict((lfn, ses) for lfn, ses in self.transReplicas.items() if lfn in runLfns)  # can be an iterator
      existingSEs = set(se for ses in replicas.values() for se in ses if se not in keepSEs)  # can be an iterator
      destinationSE = self.util.getSEForDestination(runID, existingSEs)
      if destinationSE is None:
        # If there is no replica at destination, remove randomly
        self.util.logInfo("No replicas found at destination for %d files of run %d" % (len(runLfns), runID))
        replicasWithKeep = {}
        replicasNoKeep = replicas
      else:
        self.util.logVerbose("Preparing tasks for run %d, destination %s" % (runID, destinationSE))
        replicasWithKeep = dict((lfn, ses) for lfn, ses in replicas.items()
                                if destinationSE in ses)  # can be an iterator
        replicasNoKeep = dict((lfn, ses) for lfn, ses in replicas.items()
                              if destinationSE not in ses)  # can be an iterator
      # We keep one more replica @ destinationSE, therefore decrease the number to be kept
      for reps, keep, kSEs in ((replicasNoKeep,
                                minKeep,
                                keepSEs),
                               (replicasWithKeep,
                                (abs(minKeep) - 1) * minKeep // abs(minKeep),
                                keepSEs + [destinationSE])):
        if reps:
          res = self._removeReplicas(replicas=reps, fromSEs=fromSEs, keepSEs=kSEs, minKeep=keep)
          if not res['OK']:
            self.util.logError("Error creating tasks", res['Message'])
          elif res['Value']:
            tasks += res['Value']
            targetSEs = set(se for targets, _lfns in res['Value'] for se in targets.split(','))
            runTargets = runSites.get(runID, set())
            if targetSEs - runTargets:
              # Set destination sites for that run
              runTargets = ','.join(sorted(targetSEs | runTargets))
              self.util.logVerbose("Setting destination for run %d to %s" % (runID, runTargets))
              res = self.transClient.setTransformationRunsSite(self.transID, runID, runTargets)
              if not res['OK']:
                self.util.logError("Failed to set target SEs to run %d as %s" %
                                   (runID, runTargets), res['Message'])
    return S_OK(tasks)

  def _RemoveDatasetFromDisk(self):
    """Plugin used to remove disk replicas, keeping some (e.g. archives)"""
    keepSEs = resolveSEGroup(self.util.getPluginParam('KeepSEs', ['Tier1-Archive']))
    self.util.logInfo("Starting execution of plugin")
    return self._removeReplicas(keepSEs=keepSEs, minKeep=0)

  def _RemoveReplicas(self, minKeep=None):
    """Plugin for removing replicas from specific SEs specified in FromSEs."""
    fromSEs = resolveSEGroup(self.util.getPluginParam('FromSEs', []))
    keepSEs = resolveSEGroup(self.util.getPluginParam('KeepSEs', ['Tier1-Archive']))
    mandatorySEs = resolveSEGroup(self.util.getPluginParam('MandatorySEs', []))
    # Allow removing explicitly from SEs in mandatorySEs
    mandatorySEs = [se for se in mandatorySEs if se not in fromSEs]
    # this is the minimum number of replicas to be kept in addition to keepSEs and mandatorySEs
    if minKeep is None:
      minKeep = abs(self.util.getPluginParam('NumberOfReplicas', 1))

    self.util.logInfo("Starting execution of plugin")
    return self._removeReplicas(fromSEs=fromSEs, keepSEs=keepSEs, mandatorySEs=mandatorySEs, minKeep=minKeep)

  def _ReduceReplicas(self):
    """Plugin for reducing the number of replicas to NumberOfReplicas."""
    # this is the number of replicas to be kept in addition to keepSEs and mandatorySEs
    self.util.logInfo("Starting execution of plugin")
    minKeep = -abs(self.util.getPluginParam('NumberOfReplicas', 1))
    return self._RemoveReplicas(minKeep=minKeep)

  def _removeReplicas(self, replicas=None, fromSEs=None, keepSEs=None, mandatorySEs=None, minKeep=999):
    """Utility actually implementing the logic to remove replicas or files."""
    if fromSEs is None:
      fromSEs = []
    if keepSEs is None:
      keepSEs = []
    if mandatorySEs is None:
      mandatorySEs = []
    reduceSEs = minKeep < 0
    minKeep = abs(minKeep)
    if replicas is None:
      replicas = self.transReplicas

    storageElementGroups = {}
    notInKeepSEs = []
    for replicaSE, lfns in getFileGroups(replicas).items():  # can be an iterator
      replicaSE = replicaSE.split(',')
      if minKeep == 0 and keepSEs:
        # Check that the dataset exists at least at 1 keepSE
        if not [se for se in replicaSE if se in keepSEs]:
          notInKeepSEs.extend(lfns)
          continue
      existingSEs = [se for se in replicaSE if se not in keepSEs and not self.util.dmsHelper.isSEFailover(se)]
      if minKeep == 0:
        # We only keep the replicas in keepSEs
        targetSEs = sorted(existingSEs)
      else:
        targetSEs = []
        # Take into account the mandatory SEs
        existingSEs = [se for se in existingSEs if se not in mandatorySEs]
        self.util.logVerbose("%d files, non-keep SEs: %s, removal from %s, keep %d" % (len(lfns), existingSEs,
                                                                                       fromSEs, minKeep))
        # print existingSEs, fromSEs, minKeep
        fromSet = set(fromSEs)
        existingSet = set(existingSEs)
        if len(existingSEs) > minKeep:
          # explicit deletion
          if fromSEs and not reduceSEs:
            # check how  many replicas would be left if we remove all from fromSEs
            nLeft = len(existingSet - fromSet)
            # we can delete all replicas in fromSEs
            targetSEs = list(existingSet & fromSet)
            self.util.logVerbose("Target SEs, 1st level: %s, number of left replicas: %d" % (targetSEs, nLeft))
            if nLeft < minKeep:
              # we should keep some in fromSEs, too bad
              targetSEs = randomize(targetSEs)[0:minKeep - nLeft]
              self.util.logInfo("Found %d files that could only be deleted in %d of the requested SEs" %
                                (len(lfns), minKeep - nLeft))
              self.util.logVerbose("Target SEs, 2nd level: %s" % targetSEs)
          elif fromSEs:
            # Here the fromSEs are only a preference (we want to keep only exactly minKeep replicas)
            targetSEs = randomize(list(existingSet & fromSet)) + randomize(list(existingSet - fromSet))
            targetSEs = targetSEs[0:-minKeep]
          else:
            # remove all replicas and keep only minKeep
            targetSEs = randomize(existingSEs)[0:-minKeep]
          targetSEs.sort()
          self.util.logVerbose("Remove %d replicas from %s" % (len(lfns), ','.join(targetSEs)))
        elif not reduceSEs and (existingSet & fromSet):
          nLeft = len(existingSet - fromSet)
          self.util.logInfo("Found %d files at %s with not enough replicas (%d left, %d requested), set Problematic" %
                            (len(lfns), ','.join(existingSEs), nLeft, minKeep))
          self.transClient.setFileStatusForTransformation(self.transID, 'Problematic', lfns)
          continue

      if targetSEs:
        stringTargetSEs = ','.join(targetSEs)
        storageElementGroups.setdefault(stringTargetSEs, []).extend(lfns)
      else:
        self.util.logInfo("Found %s files that don't need any replica deletion, set Processed" % len(lfns))
        self.transClient.setFileStatusForTransformation(self.transID, 'Processed', lfns)

    if notInKeepSEs:
      self.util.logInfo("Found %d files not in at least one keepSE, no removal done, set Problematic" %
                        len(notInKeepSEs))
      self.transClient.setFileStatusForTransformation(self.transID, 'Problematic', notInKeepSEs)

    if self.pluginCallback:
      self.pluginCallback(self.transID, invalidateCache=True)
    return S_OK(self.util.createTasks(storageElementGroups))

  def _RemoveReplicasWhenProcessed(self, maxFiles=None):
    """This plugin considers files and checks whether they were processed for a
    list of processing passes For files that were processed, it sets replica
    removal tasks from a set of SEs."""
    keepSEs = resolveSEGroup(self.util.getPluginParam('KeepSEs', []))
    fromSEs = set(resolveSEGroup(self.util.getPluginParam('FromSEs', []))) - set(keepSEs)
    # Ignore files that are at a banned SE
    bannedSEs = set(se for se in fromSEs if not StorageElement(se).status()['Remove'])
    if not fromSEs:
      self.util.logError('No SEs where to delete from, check overlap with %s' % keepSEs)
      return S_OK([])
    processingPasses = self.util.getPluginParam('ProcessingPasses', [])

    transStatus = self.params['Status']
    self.util.readCacheFile(self.workDirectory)

    if not processingPasses:
      self.util.logWarn("No processing pass(es)")
      return S_OK([])

    maxTime = self.util.getPluginParam('MaxTimeAllowed', 0)
    pluginStartTime = time.time()

    self.util.setCachedTimeExceeded(False)
    try:
      # Now we must find out whether the input files have a descendant in the processing passes
      result = self.util.getProductions(processingPasses, transStatus)
      if not result['OK']:
        return result
      if not result['Value']:
        return S_OK([])
      bkPathList, productions = result['Value']
      if productions is None or not productions.get('List'):
        return S_OK([])

      # Group files per StorageElement
      replicaGroups = getFileGroups(self.transReplicas)
      self.util.logVerbose("Using %d input files, in %d groups" % (len(self.transReplicas), len(replicaGroups)))
      storageElementGroups = {}
      newGroups = {}
      for stringSEs, lfns in replicaGroups.items():  # can be an iterator
        replicaSEs = set(stringSEs.split(','))
        if replicaSEs & bannedSEs:
          # Ignore these files
          continue
        targetSEs = fromSEs & replicaSEs
        if not targetSEs:
          # This is a fake to have a placeholder for the replica location... Later it is not used
          self.util.logVerbose("%d files are not in required list (only at %s)" %
                               (len(lfns), ','.join(sorted(replicaSEs))))
          newGroups.setdefault(','.join(sorted(replicaSEs)), []).extend(lfns)
        elif not replicaSEs - fromSEs:
          self.util.logInfo("%d files are only in required list (only at %s), don't remove (yet)" %
                            (len(lfns), ','.join(sorted(replicaSEs))))
        else:
          self.util.logVerbose("%d files are in required list (also at %s)" %
                               (len(lfns), ','.join(sorted(replicaSEs - fromSEs))))
          newGroups.setdefault(','.join(sorted(targetSEs)), []).extend(lfns)

      # Restrict the query to the BK to the interesting productions
      #####################
      # Loop on storages  #
      #####################
      for stringTargetSEs in randomize(newGroups):
        # if enough time already spent, exit
        # this is better placed at the beginning of the loop in case timeout in last item
        timeSpent = time.time() - pluginStartTime
        if maxTime and timeSpent > maxTime:
          # We break the loop here and set a flag in the cached file to execute next time
          self.util.setCachedTimeExceeded(True)
          self.util.logInfo("Enough time spent in plugin (%.1f seconds), exit" % (timeSpent))
          break

        lfns = set(newGroups[stringTargetSEs])
        # Use the cached information if any
        bkPathsToCheck = dict((lfn, set(self.util.cachedLFNProcessedPath.get(lfn, bkPathList)) & set(bkPathList))
                              for lfn in lfns)
        # Only check files that are not fully processed (cache contains processing passes still not done)
        lfnsToCheck = set(lfn for lfn in bkPathsToCheck if bkPathsToCheck[lfn])
        self.util.logInfo('Checking descendants for %d files at %s' % (len(lfnsToCheck), stringTargetSEs))
        # Update with the cached information
        remaining = len(bkPathList)
        for bkPath in bkPathList:
          # How many iterations are left?
          remaining -= 1
          prods = productions['List'][bkPath]
          # If there is nothing left to do, exit
          if not prods or not lfnsToCheck:
            break
          lfnsToCheckForPath = set(lfn for lfn in lfnsToCheck if bkPath in bkPathsToCheck[lfn])
          if not lfnsToCheckForPath:
            continue
          startTime = time.time()
          res = self.util.checkForDescendants(lfnsToCheckForPath, prods)
          if not res['OK']:
            self.util.logError("Error checking descendants using utility", res['Message'])
          processedLfns = res.get('Value', set())
          self.util.logVerbose('Found %s processed files in %.1f seconds' %
                               (len(processedLfns) if processedLfns else 'no',
                                time.time() - startTime))
          # Remove bkPath from processing passes to check for lLFNs found processed
          for lfn in processedLfns:
            if bkPath not in bkPathsToCheck[lfn]:
              self.util.logWarn('LFN not in list: %s' % lfn, str(bkPathsToCheck[lfn]))
            else:
              bkPathsToCheck[lfn].remove(bkPath)
          lfnsToCheckForPath -= processedLfns
          # Only worth checking if not the last processing pass
          if remaining:
            notProcessed = set(lfn for lfn in lfnsToCheckForPath if bkPath in bkPathsToCheck[lfn])
            if notProcessed:
              self.util.logVerbose("%d files not processed by processing pass %s, don't check further" %
                                   (len(notProcessed), bkPathList[bkPath]))
              lfnsToCheck -= notProcessed

        lfnsProcessed = [lfn for lfn in lfns if not bkPathsToCheck[lfn]]
        self.util.cachedLFNProcessedPath.update(bkPathsToCheck)
        # print lfnsProcessed, bkPathsToCheck
        self.util.logInfo("Found %d / %d files that are processed (/ not) at %s" %
                          (len(lfnsProcessed),
                           len([lfn for lfn in lfns if bkPathsToCheck[lfn]]),
                           stringTargetSEs))
        if lfnsProcessed:
          targetSEs = set(stringTargetSEs.split(','))
          if not targetSEs & fromSEs:
            # Files are processed but are no longer at the requested SEs, set them Processed
            self._alreadyProcessedLFNs.setdefault(','.join(fromSEs), []).extend(lfnsProcessed)
            self.util.logInfo("%d processed files are no longer in required SE list: set them Processed" %
                              len(lfnsProcessed))
            self.transClient.setFileStatusForTransformation(self.transID, 'Processed', lfnsProcessed)
          else:
            storageElementGroups.setdefault(stringTargetSEs, []).extend(lfnsProcessed)

        ###############
        # End of loop #
        ###############
      if not storageElementGroups:
        return S_OK([])
    except Exception as e:  # pylint: disable=broad-except
      self.util.logException('Exception while executing the plugin', '', lException=e)
      return S_ERROR(e)
    finally:
      self.util.writeCacheFile()
      if self.pluginCallback:
        self.pluginCallback(self.transID, invalidateCache=True)
    return S_OK(self.util.createTasks(storageElementGroups, chunkSize=maxFiles))

  def _RemoveReplicasWithAncestors(self):
    """Same as _RemoveReplicasWhenProcessed but also remove parents This plugin
    is useful for removing at once RDST and RAW files after stripping."""
    return self.__addAncestors(pluginMethod=self._RemoveReplicasWhenProcessed)

  def __getAncestorLFNs(self, lfns):
    ancestors = self.util.getFileAncestors(lfns, depth=1)
    if not ancestors['OK']:
      self.util.logError("Error getting ancestors", ancestors['Message'])
      return ancestors
    ancestors = [anc['FileName'] for ancList in ancestors['Value']['Successful'].values()
                 for anc in ancList]  # can be an iterator
    return S_OK(ancestors)

  def __addAncestors(self, pluginMethod=None):
    """Call a standard plugin and then add ancestors to tasks."""
    maxFiles = self.util.getPluginParam('MaxFilesPerTask', 100) // 2
    tasks = pluginMethod(maxFiles=maxFiles)
    if not tasks['OK']:
      return tasks
    newTasks = []
    addedAncestors = []
    for targetSE, lfns in tasks['Value']:
      ancestors = self.__getAncestorLFNs(lfns)
      if not ancestors['OK']:
        return ancestors
      ancestors = ancestors['Value']
      # It is not possible to create a task with files that are not in the transformation!
      # therefore add them...
      if ancestors:
        res = self.transClient.addFilesToTransformation(self.transID, ancestors)
        if not res['OK']:
          self.util.logError('Failed to add files to transformation', res['Message'])
          return res
        # Only put added files in tasks
        addedLfns = [lfn for (lfn, status) in res['Value']['Successful'].items()
                     if status == 'Added']  # can be an iterator
        addedAncestors += addedLfns
      else:
        addedLfns = []
      newTasks.append((targetSE, lfns + addedLfns))
    # This dict is those files already processed by the initial plugin (i.e. no need to process them)
    for targetSE, lfns in self._alreadyProcessedLFNs.items():  # can be an iterator
      ancestors = self.__getAncestorLFNs(lfns)
      if not ancestors['OK']:
        return ancestors
      ancestors = ancestors['Value']
      if ancestors:
        res = self.transClient.addFilesToTransformation(self.transID, ancestors)
        if not res['OK']:
          self.util.logError('Failed to add files to transformation', res['Message'])
          return res
        addedLfns = [lfn for (lfn, status) in res['Value']['Successful'].items()
                     if status == 'Added']  # can be an iterator
        self.util.logVerbose("Found %d ancestors of Processed files: add them to tasks" % len(addedLfns))
        addedAncestors += addedLfns
        for ancChunk in breakListIntoChunks(addedLfns, 2 * maxFiles):
          newTasks.append((targetSE, ancChunk))
    if addedAncestors:
      self.util.logInfo("Added %d ancestors to tasks" % len(addedAncestors))
    return S_OK(newTasks)

  def _ReplicateToLocalSE(self, maxFiles=None):
    """Used for example to replicate from a buffer to a tape SE on the same
    site."""
    res = self.util.getMaxFilesToReplicate(self.workDirectory)
    if not res['OK']:
      return res
    watermark, maxFilesAtSE = res['Value']
    # This is a convention, to skip the loop or to stop replicating
    if watermark is None:
      return S_OK([])
    destSEs = set(maxFilesAtSE)
    storageElementGroups = {}

    for replicaSE, lfns in getFileGroups(self.transReplicas).items():  # can be an iterator
      replicaSEs = set(se for se in replicaSE.split(',') if not self.util.dmsHelper.isSEFailover(se))
      if not replicaSEs:
        continue
      okSEs = replicaSEs & destSEs
      if okSEs:
        # We have to choose only one SE to replicate to...
        #  There should in principle not be more but to be safe, take one only
        self._alreadyProcessedLFNs.setdefault(list(okSEs)[0], []).extend(lfns)
        self.util.logInfo("Found %d files that are already present in the destination SEs (status set Processed)" %
                          len(lfns))
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Processed', lfns)
        if not res['OK']:
          self.util.logError("Can't set files to Processed", '(%d files): %s' % (len(lfns), res['Message']))
          return res
        continue
      targetSEs = destSEs - replicaSEs
      candidateSEs = self.util.closerSEs(replicaSEs, targetSEs, local=True)
      if candidateSEs:
        # If the max number of files to copy is negative, stop
        shortSEs = [se for se in candidateSEs if maxFilesAtSE.get(se, sys.maxsize) == 0]
        candidateSEs = [se for se in candidateSEs if se not in shortSEs]
        if not candidateSEs:
          self.util.logInfo("No candidate SE where more files are accepted (%s not allowed)" % ','.join(shortSEs))
        else:
          # Check if enough free space
          freeSpace = self.util.getStorageFreeSpace(candidateSEs)
          shortSEs = [se for se in candidateSEs if freeSpace[se] < watermark]
          candidateSEs = [se for se in candidateSEs if se not in shortSEs]
          if not candidateSEs:
            self.util.logInfo("No enough space (%s TB) found at %s" % (watermark, ','.join(shortSEs)))
          else:
            # Select a single SE out of candidates; in most cases there is one only
            candidateSE = candidateSEs[0]
            maxToReplicate = maxFilesAtSE.get(candidateSE, sys.maxsize)
            if maxToReplicate < len(lfns):
              self.util.logInfo("Limit number of files for %s to %d (out of %d)" %
                                (candidateSE, maxToReplicate, len(lfns)))
            else:
              self.util.logVerbose("Number of files for %s: %d" % (candidateSE, len(lfns)))
            storageElementGroups.setdefault(candidateSE, []).extend(lfns[:maxToReplicate])
      else:
        self.util.logWarn("Could not find a local SE for %d files, set them Problematic" % len(lfns))
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Problematic', lfns)
        if not res['OK']:
          self.util.logError("Can't set files to Problematic", '(%d files): %s' %
                             (len(lfns), res['Message']))
          return res

    return S_OK(self.util.createTasks(storageElementGroups, chunkSize=maxFiles))

  def _ReplicateWithAncestors(self):
    """Same as _ReplicateToLocalSE but also replicate parents If only one SE is
    given, use _ReplicateDataset This plugin is useful for prestaging at once
    RDST and RAW files before stripping."""
    destSEs = set(resolveSEGroup(self.util.getPluginParam('DestinationSEs', [])))
    if len(destSEs) == 1:
      return self.__addAncestors(pluginMethod=self._ReplicateDataset)
    return self.__addAncestors(pluginMethod=self._ReplicateToLocalSE)

  def _Healing(self):
    """Plugin that creates task for replicating files to the same SE where they
    are declared problematic."""
    self.util.cleanFiles(self.transFiles, self.transReplicas)
    storageElementGroups = {}

    for replicaSE, lfns in getFileGroups(self.transReplicas).items():  # can be an iterator
      replicaSE = set(se for se in replicaSE.split(',') if not self.util.dmsHelper.isSEFailover(se))
      if not replicaSE:
        self.util.logInfo("Found %d files that don't have a suitable source replica. Set Problematic" % len(lfns))
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Problematic', lfns)
        continue
      # get no problematic replicas only
      res = self.fileCatalog.getReplicas(lfns, allStatus=False)
      if not res['OK']:
        self.util.logError('Error getting catalog replicas', res['Message'])
        continue
      replicas = res['Value']['Successful']
      noMissingSE = []
      noOtherReplica = []
      for lfn in lfns:
        if lfn not in replicas:
          # This file has no active replicas, problematic
          noOtherReplica.append(lfn)
        else:
          targetSEs = replicaSE - set(replicas[lfn])
          if targetSEs:
            storageElementGroups.setdefault(','.join(sorted(targetSEs)), []).append(lfn)
          else:
            # print lfn, sorted( replicas[lfn] ), sorted( replicaSE )
            noMissingSE.append(lfn)
      if noOtherReplica:
        self.util.logInfo("Found %d files that have no other active replica (set Problematic)" % len(noOtherReplica))
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Problematic', noOtherReplica)
        if not res['OK']:
          self.util.logError("Can't set %d files of transformation %s to 'Problematic': %s" % (len(noOtherReplica),
                                                                                               str(self.transID),
                                                                                               res['Message']))
      if noMissingSE:
        self.util.logInfo(
            "Found %d files that are already present in the destination SEs (set Processed)" %
            len(noMissingSE))
        res = self.transClient.setFileStatusForTransformation(self.transID, 'Processed', noMissingSE)
        if not res['OK']:
          self.util.logError("Can't set %d files of transformation %s to 'Processed': %s" % (len(noMissingSE),
                                                                                             str(self.transID),
                                                                                             res['Message']))
          return res

    return S_OK(self.util.createTasks(storageElementGroups))



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
  Execute a (list of) commands taking arguments from a file.
  Usage:
      dirac-loop [File] Command [Command2...] [option|cfgfile]

  Arguments:
     File: File containing the strings to be appended to the command (default /dev/stdin)
           If the arguments are BK paths, the script reduces the paths merging event and file types
           (use --NoMerge to disable)
     Command: Command to execute (separate with spaces)
              If command terminates with ", a " is appended after the argument
              If the command contains the string @arg@,
                it is replaced by the argument rather than the argument appended
              If the command itself is dirac-loop, then the first occurrence only of @arg@ is replaced,
              which allows recursive calls
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import subprocess
from collections import defaultdict


from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC.Core.Utilities.List import breakListIntoChunks


def reduceArgs(noMerge, arguments, maxList=20):
  """
  If the arguments look like BK paths (start with /LHCb or /MC), try to
  reduce the list of BK paths by merging event types or file types into a
  list.

  :param bool: if set, do not merge arguments if BK paths
  :param list arguments: list of arguments
  :param int maxList: maximum number of event/file types to put in a resulting merged argument
  """
  if noMerge:
    return arguments
  others = []
  placeHolder = '@@'
  for item in [-1, -2]:
    reducedDict = defaultdict(set)
    for path in arguments:
      path = path.strip()
      if path.startswith('/LHCb') or path.startswith('/MC'):
        parsed = path.split('/')
        val = parsed[item]
        parsed[item] = placeHolder
        holderPath = '/'.join(parsed)
        reducedDict[holderPath].add(val)
      else:
        others.append(path)
    # Don't allow list of event or file types too long
    arguments = []
    for holderPath, values in reducedDict.items():
      for chunk in breakListIntoChunks(sorted(values), maxList):
        arguments.append(holderPath.replace(placeHolder, ','.join(chunk)))
  # Now sort out the conditions
  conditions = {}
  finalArgs = [arg for arg in arguments if not arg.startswith('/LHCb')]
  for path in [arg for arg in arguments if arg.startswith('/LHCb')]:
    parsed = path.split('/')
    cond = parsed[3]
    parsed[3] = ''
    conditions.setdefault('/'.join(parsed), []).append((path, cond))
  for newPath, condTuple in conditions.items():
    if len(condTuple) != 2:
      finalArgs += [path for (path, _c) in condTuple]
    else:
      if len(set(['-'.join(cond.split('-')[0:2])])) == 1:
        finalArgs.append(newPath)
      else:
        finalArgs += [path for (path, _c) in condTuple]

  return sorted(others + finalArgs)


@DIRACScript()
def main():
  from DIRAC import gLogger
  from DIRAC.Core.Base import Script

  Script.registerSwitch('', 'NoMerge', 'If set, do not merge arguments if BK paths')
  Script.registerSwitch('', 'Items=', 'Alternative way of passing list of arguments')
  Script.registerSwitch('', 'NoParse', 'Consider the full lines are arguments and not only the first word')
  Script.registerSwitch('', 'Terse', 'Do not print the command executed')
  Script.setUsageMessage('\n'.join([__doc__]))
  Script.parseCommandLine(ignoreErrors=True)
  args = Script.getPositionalArgs()

  noMerge = False
  noParse = False
  terse = False
  arguments = []
  for switch, val in Script.getUnprocessedSwitches():
    if switch == 'NoMerge':
      noMerge = True
    elif switch == 'NoParse':
      noParse = True
    elif switch == 'Items':
      arguments = val.split(',')
    elif switch == 'Terse':
      terse = True

  if len(args) < 1:
    Script.showHelp(exitCode=1)

  if not arguments:
    if os.path.exists(args[0]):
      oFile = args.pop(0)
    else:
      oFile = '/dev/stdin'
    with open(oFile, 'r') as fd:
      arguments = fd.read().split('\n')[:-1]

  commands = args

  argList = []
  for arg in arguments:
    if not arg.strip():
      continue
    # If the argument is between quotes, take what is between the quotes
    if arg[0] == "'":
      arg = arg.split("'")[1]
    elif arg[0] == '"':
      arg = arg.split('"')[1]
    elif noParse:
      # Consider the whole line is the argument
      pass
    else:
      # Take the firt word
      arg = arg.strip().split()[0]
    # Escape any space left
    argList.append(arg.replace(' ', r'\ '))

  for arg in reduceArgs(noMerge, argList):
    if arg:
      for command in commands:
        if '@arg@' in command:
          c = command.replace('@arg@', arg, 1) if 'dirac-loop' in command else \
              command.replace('@arg@', arg)
        elif command[-1] in ('"', "'"):
          c = command + arg.replace(r'\ ', ' ') + command[-1]
        else:
          c = command + ' ' + arg
        if not terse:
          gLogger.notice('======= %s =========' % c)
        try:
          output = subprocess.check_output(c, stderr=subprocess.STDOUT, shell=True)
          gLogger.notice(output[:-1] if terse else output)
        except subprocess.CalledProcessError as e:
          gLogger.error("Error calling command, return code %d\n" % e.returncode, e.output)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Gets all Assigned files in a transformation and reports by target SE."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def __getTransformations(args):
  from DIRAC.Core.Base import Script
  if not len(args):
    print("Specify transformation number...")
    Script.showHelp()
  else:
    ids = args[0].split(",")
    transList = []
    for transID in ids:
      r = transID.split(':')
      if len(r) > 1:
        for i in range(int(r[0]), int(r[1]) + 1):
          transList.append(i)
      else:
        transList.append(int(r[0]))
  return transList


def __getTask(transClient, transID, taskID):
  res = transClient.getTransformationTasks({'TransformationID': transID, "TaskID": taskID})
  if not res['OK'] or not res['Value']:
    return None
  return res['Value'][0]


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  Script.parseCommandLine(ignoreErrors=True)
  transList = __getTransformations(Script.getPositionalArgs())

  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  from DIRAC import gLogger
  transClient = TransformationClient()

  for transID in transList:
    res = transClient.getTransformationFiles({'TransformationID': transID, 'Status': 'Assigned'})
    if not res['OK']:
      gLogger.fatal("Error getting transformation files for %d" % transID)
      continue
    targetStats = {}
    taskDict = {}
    for fileDict in res['Value']:
      taskID = fileDict['TaskID']
      taskDict[taskID] = taskDict.setdefault(taskID, 0) + 1
    for taskID in taskDict:
      task = __getTask(transClient, transID, taskID)
      targetSE = task.get('TargetSE', None)
      targetStats[targetSE][0] = targetStats.setdefault(targetSE, [0, 0])[0] + taskDict[taskID]
      targetStats[targetSE][1] += 1

    gLogger.always("Transformation %d: %d assigned files found" % (transID, len(res['Value'])))
    for targetSE, (nfiles, ntasks) in targetStats.items():
      gLogger.always("\t%s: %d files in %d tasks" % (targetSE, nfiles, ntasks))


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This script allows to print information about a (list of)
transformations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  informations = ['AuthorDN', 'AuthorGroup', 'Body', 'CreationDate',
                  'Description', 'EventsPerTask', 'FileMask', 'GroupSize', 'Hot',
                  'InheritedFrom', 'LastUpdate', 'LongDescription', 'MaxNumberOfTasks',
                  'Plugin', 'Status', 'TransformationGroup',
                  'TransformationName', 'Type', 'Request']
  Script.registerSwitch('', 'Information=', '   Specify which information is required')
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [options] transID1 [transID2 ...]' % Script.scriptName,
                                    'Arguments:',
                                    '\ttransID1,... : transformantion IDs',
                                    'Possible informations:',
                                    '\t%s' % ', '.join(sorted(informations))
                                    ])
                         )
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  from DIRAC import gLogger

  tr = TransformationClient()

  requestedInfo = informations
  switches = Script.getUnprocessedSwitches()
  infoList = []
  for switch, val in switches:
    if switch == 'Information':
      infoList = [info.lower() for info in val.split(',')]
      requestedInfo = [info for info in informations if info.lower() in infoList]
  if 'body' not in infoList and 'Body' in requestedInfo:
    requestedInfo.remove('Body')

  transIDs = Script.getPositionalArgs()

  for transID in transIDs:
    try:
      res = tr.getTransformation(int(transID))
      gLogger.notice("==== Transformation %s ====" % transID)
      for info in requestedInfo:
        getInfo = info if info != 'Request' else 'TransformationFamily'
        gLogger.notice("\t%s: %s" %
                       (info, res.get('Value', {}).get(getInfo, 'Unknown')))
    except Exception:
      gLogger.error("Invalid transformation ID: '%s'" % transID)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""BKQuery is a class that decodes BK paths, queries the BK at a high level."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import sys
import six
from fnmatch import fnmatch
from DIRAC import gLogger
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


def getProcessingPasses(bkQuery, depth=None):
  """Get the list of processing passes for a given BK query The processing pass
  in the initial query may contain "..." or a '*', in which case this acts as a
  wildcard character The search for processing passes may be limited to a
  certain depth (default: all)

  :param bkQuery: BK query dictionary
  :type bkQuery: dict
  :param depth: depth to which look for the processing pass
  :type depth: int
  """
  processingPass = bkQuery.getProcessingPass().replace('...', '*')
  if '*' not in processingPass:
    return [processingPass]
  ind = processingPass.index('*')
  basePass = os.path.dirname(processingPass[:ind])
  bkQuery.setProcessingPass(basePass)
  return sorted(pp for pp in bkQuery.getBKProcessingPasses(depth=depth)
                if fnmatch(pp, processingPass) and pp != basePass)


def makeBKPath(bkDict):
  """Builds a path from the dictionary.

  :param bkDict: BK query dictionary
  :type bkDict: dict
  """
  fileType = bkDict.get('FileType', '.')
  if isinstance(fileType, list):
    fileType = ','.join(fileType)
  path = os.path.join('/',
                      bkDict.get('ConfigName', ''),
                      bkDict.get('ConfigVersion', ''),
                      bkDict.get('ConditionDescription',
                                 bkDict.get('DataTakingConditions',
                                            bkDict.get('SimulationConditions', '.'))),
                      bkDict.get('ProcessingPass', '.')[1:],
                      str(bkDict.get('EventType', '.')).replace('90000000', '.'),
                      fileType).replace('/.', '/')
  while path.endswith('/'):
    path = path[:-1]
  return path.replace('RealData', 'Real Data')


class BadRunRange(Exception):
  """Exception class for bad run range."""
  pass


def parseRuns(bkQuery, runs):
  """Parse the parameter "Runs" and set it in the dictionary.

  :param bkQuery: BK dictionary
  :type bkQuery: dict
  :param runs: a string, an int, or a list,dict,tuple of run numbers
  :type runs: string or int/long, iterable
  """
  if isinstance(runs, six.string_types):
    runs = runs.split(',')
  elif isinstance(runs, (dict, tuple)):
    runs = list(runs)
  elif isinstance(runs, six.integer_types):
    runs = [str(runs)]
  if len(runs) > 1:
    runList = []
    for run in runs:
      if run.isdigit():
        runList.append(int(run))
      else:
        runRange = run.split(':')
        if len(runRange) == 2 and runRange[0].isdigit() and runRange[1].isdigit():
          runList += range(int(runRange[0]), int(runRange[1]) + 1)
        else:
          gLogger.error("Run numbers must be numbers...")
          raise BadRunRange
    bkQuery['RunNumber'] = runList
  else:
    runs = runs[0].split(':')
    if len(runs) == 1:
      runs = runs[0].split('-')
      if len(runs) == 1:
        bkQuery['RunNumber'] = int(runs[0])
  if 'RunNumber' not in bkQuery:
    try:
      if runs[0] and runs[1] and int(runs[0]) > int(runs[1]):
        gLogger.error('Warning: End run should be larger than start run: %d, %d' % (int(runs[0]),
                                                                                    int(runs[1])))
        raise BadRunRange
      if runs[0].isdigit():
        bkQuery['StartRun'] = int(runs[0])
      if runs[1].isdigit():
        bkQuery['EndRun'] = int(runs[1])
    except IndexError as ex:  # The runs must be a list
      gLogger.exception("Invalid run range", runs, lException=ex)
      raise BadRunRange
  else:
    if 'StartRun' in bkQuery:
      bkQuery.pop('StartRun')
    if 'EndRun' in bkQuery:
      bkQuery.pop('EndRun')
  return bkQuery


class BKQuery():
  """It used to build a dictionary using a given Bookkeeping path which is used
  to query the Bookkeeping database."""

  def __init__(self, bkQuery=None, prods=None, runs=None, fileTypes=None, visible=True, eventTypes=None):
    prods = prods if prods is not None else []
    runs = runs if runs is not None else []
    fileTypes = fileTypes if fileTypes is not None else []

    self.extraBKitems = ("StartRun", "EndRun", "Production", "RunNumber")
    self.__bkClient = BookkeepingClient()
    bkPath = ''
    bkQueryDict = {}
    self.__bkFileTypes = set()
    self.__exceptFileTypes = set()
    self.__fakeAllDST = 'ZZZZZZZZALL.DST'
    self.__alreadyWarned = False
    if isinstance(bkQuery, BKQuery):
      bkQueryDict = bkQuery.getQueryDict().copy()
    elif isinstance(bkQuery, dict):
      bkQueryDict = bkQuery.copy()
    elif isinstance(bkQuery, six.string_types):
      bkPath = bkQuery
    bkQueryDict = self.buildBKQuery(bkPath=bkPath, bkQueryDict=bkQueryDict,
                                    prods=prods, runs=runs,
                                    fileTypes=fileTypes, eventTypes=eventTypes,
                                    visible=visible)
    self.__bkPath = bkPath
    self.__bkQueryDict = bkQueryDict
    if not bkQueryDict.get('Visible'):
      self.setVisible(visible)

  def __str__(self):
    return str(self.__bkQueryDict)

  def buildBKQuery(self, bkPath='', bkQueryDict=None, prods=None, runs=None,
                   fileTypes=None, visible=True, eventTypes=None):
    """it builds a dictionary using a path."""
    bkQueryDict = bkQueryDict if bkQueryDict is not None else {}
    prods = prods if prods is not None else []
    if not isinstance(prods, list):
      prods = [prods]
    runs = runs if runs is not None else []
    fileTypes = fileTypes if fileTypes is not None else []

    gLogger.verbose("BKQUERY.buildBKQuery: Path %s, Dict %s, \
      Prods %s, Runs %s, FileTypes %s, EventTypes %s, Visible %s" % (bkPath,
                                                                     str(bkQueryDict),
                                                                     str(prods),
                                                                     str(runs),
                                                                     str(fileTypes),
                                                                     str(eventTypes),
                                                                     visible))
    self.__bkQueryDict = {}
    if not bkPath and not prods and not bkQueryDict and not runs:
      return {}
    if bkQueryDict:
      bkQuery = bkQueryDict.copy()
    else:
      bkQuery = {}

    # Query given as a path /ConfigName/ConfigVersion/ConditionDescription/ProcessingPass/EventType/FileType
    # or if prefixed with evt: /ConfigName/ConfigVersion/EventType/ConditionDescription/ProcessingPass/FileType
    if bkPath:
      self.__getAllBKFileTypes()
      bkFields = ("ConfigName", "ConfigVersion", "ConditionDescription", "ProcessingPass", "EventType", "FileType")
      url = bkPath.split(':', 1)
      if len(url) == 1:
        bkPath = url[0]
      else:
        if url[0] in ('evt', 'evt+std'):
          bkFields = ("ConfigName", "ConfigVersion",
                      "EventType", "ConditionDescription",
                      "ProcessingPass", "FileType")
        elif url[0] == 'pp':
          bkFields = ("ProcessingPass", "EventType", "FileType")
        elif url[0] == 'prod':
          bkFields = ("Production", "ProcessingPass", "EventType", "FileType")
        elif url[0] == 'runs':
          bkFields = ("Runs", "ProcessingPass", "EventType", "FileType")
        elif url[0] not in ('sim', 'daq', 'cond'):
          gLogger.error('Invalid BK path:%s' % bkPath)
          return self.__bkQueryDict
        bkPath = url[1]
      if bkPath[0] != '/':
        bkPath = '/' + bkPath
      if bkPath[0:2] == '//':
        bkPath = bkPath[1:]
      bkPath = bkPath.replace("RealData", "Real Data")
      i = 0
      processingPass = '/'
      defaultPP = False
      bk = bkPath.split('/')[1:] + len(bkFields) * ['']
      for bpath in bk:
        gLogger.verbose('buildBKQuery.1. Item #%d, Field %s, From Path %s, ProcessingPass %s' % (i,
                                                                                                 bkFields[i],
                                                                                                 bpath,
                                                                                                 processingPass))
        if bkFields[i] == 'ProcessingPass':
          if bpath != '' and bpath.upper() != 'ALL' and \
              not bpath.split(',')[0].split(' ')[0].isdigit() and \
                  not bpath.upper() in self.__bkFileTypes:
            processingPass = os.path.join(processingPass, bpath)
            continue
          # Set the PP
          if processingPass != '/':
            bkQuery['ProcessingPass'] = processingPass
          else:
            defaultPP = True
          i += 1
        gLogger.verbose('buildBKQuery.2. Item #%d, Field %s, From Path %s, ProcessingPass %s' % (i,
                                                                                                 bkFields[i],
                                                                                                 bpath,
                                                                                                 processingPass))
        if bkFields[i] == 'EventType' and bpath:
          eventTypeList = []
          # print b
          if bpath.upper() == 'ALL':
            bpath = 'ALL'
          else:
            for et in bpath.split(','):
              try:
                eventType = int(et.split(' ')[0])
                eventTypeList.append(eventType)
              except ValueError:
                pass
            if len(eventTypeList) == 1:
              eventTypeList = eventTypeList[0]
            bpath = eventTypeList
            gLogger.verbose('buildBKQuery. Event types %s' % eventTypeList)
        # Set the BK dictionary item
        if bpath != '':
          bkQuery[bkFields[i]] = bpath
        if defaultPP:
          # PP was empty, try once more to get the Event Type
          defaultPP = False
        else:
          # Go to next item
          i += 1
        if i == len(bkFields):
          break

      gLogger.verbose('buildBKQuery. Query dict %s' % str(bkQuery))
      # Set default event type to real data
      if bkQuery.get('ConfigName') != 'MC' and not bkQuery.get('EventType'):
        bkQuery['EventType'] = '90000000'
      if bkQuery.get('EventType') == 'ALL':
        bkQuery.pop('EventType')

    # Run limits are given
    runs = bkQuery.pop('Runs', runs)
    if runs:
      try:
        bkQuery = parseRuns(bkQuery, runs)
      except BadRunRange:
        return self.__bkQueryDict

    # Query given as a list of production
    if prods and str(prods[0]).upper() != 'ALL':
      try:
        bkQuery.setdefault('Production', []).extend([int(prod) for prod in prods])
      except ValueError as ex:  # The prods list does not contains numbers
        gLogger.warn(ex)
        gLogger.error('Invalid production list', str(prods))
        return self.__bkQueryDict

    # If an event type is specified
    if eventTypes:
      bkQuery['EventType'] = eventTypes

    # Set the file type(s) taking into account excludes file types
    fileTypes = bkQuery.get('FileType', fileTypes)
    bkQuery.pop('FileType', None)
    self.__bkQueryDict = bkQuery.copy()
    fileType = self.__fileType(fileTypes)
    # print fileType
    if fileType:
      bkQuery['FileType'] = fileType

    # Remove all "ALL"'s in the dict, if any
    for i in self.__bkQueryDict:
      if isinstance(bkQuery[i], six.string_types) and bkQuery[i] == 'ALL':
        bkQuery.pop(i)

    # If there is only one production, make it faster with a single value rather than a list
    prodList = bkQuery.get('Production')
    if isinstance(prodList, list) and len(prodList) == 1:
      bkQuery['Production'] = prodList[0]
    self.__bkQueryDict = bkQuery.copy()
    self.setVisible(visible)

    # Set both event type entries
    # print "Before setEventType", self.__bkQueryDict
    if not self.setEventType(bkQuery.get('EventType')):
      self.__bkQueryDict = {}
      return self.__bkQueryDict
    # Set conditions
    # print "Before setConditions", self.__bkQueryDict
    self.setConditions(bkQuery.get('ConditionDescription',
                                   bkQuery.get('DataTakingConditions', bkQuery.get('SimulationConditions'))
                                   ))
    # print "Returned value", self.__bkQueryDict
    return self.__bkQueryDict

  def setOption(self, key, val):
    """It insert an item to the dictionary.

    The key is an bookkeeping attribute (condition).
    """
    if val:
      self.__bkQueryDict[key] = val
    else:
      self.__bkQueryDict.pop(key, None)
    return self.__bkQueryDict

  def setConditions(self, cond=None):
    """Set the dictionary items for a given condition, or remove it
    (cond=None)"""
    if 'ConfigName' not in self.__bkQueryDict and cond:
      gLogger.warn("Impossible to set Conditions to a BK Query without Configuration")
      return self.__bkQueryDict
    # There are two items in the dictionary: ConditionDescription and Simulation/DataTaking-Conditions
    eventType = self.__bkQueryDict.get('EventType', 'ALL')
    if self.__bkQueryDict.get('ConfigName') == 'MC' or \
        (isinstance(eventType, six.string_types) and eventType.upper() != 'ALL'
         and eventType[0] != '9'):
      conditionsKey = 'SimulationConditions'
    else:
      conditionsKey = 'DataTakingConditions'
    self.setOption('ConditionDescription', cond)
    return self.setOption(conditionsKey, cond)

  def setFileType(self, fileTypes=None):
    """insert the file type to the Boookkeeping dictionary."""
    return self.setOption('FileType', self.__fileType(fileTypes))

  def setDQFlag(self, dqFlag='OK'):
    """Sets the data quality."""
    if isinstance(dqFlag, six.string_types):
      dqFlag = dqFlag.upper()
    elif isinstance(dqFlag, list):
      dqFlag = [dq.upper() for dq in dqFlag]
    return self.setOption('DataQuality', dqFlag)

  def setStartDate(self, startDate):
    """Sets the start date."""
    return self.setOption('StartDate', startDate)

  def setEndDate(self, endDate):
    """Sets the end date."""
    return self.setOption('EndDate', endDate)

  def setProcessingPass(self, processingPass):
    """Sets the processing pass."""
    return self.setOption('ProcessingPass', processingPass)

  def setEventType(self, eventTypes=None):
    """Sets the event type."""
    if eventTypes:
      if isinstance(eventTypes, six.string_types):
        eventTypes = eventTypes.split(',')
      elif not isinstance(eventTypes, list):
        eventTypes = [eventTypes]
      try:
        eventTypes = [str(int(et)) for et in eventTypes]
      except ValueError as ex:
        gLogger.warn(ex)
        gLogger.error('Invalid list of event types', eventTypes)
        return {}
      if isinstance(eventTypes, list) and len(eventTypes) == 1:
        eventTypes = eventTypes[0]
    return self.setOption('EventType', eventTypes)

  def setVisible(self, visible=None):
    """Sets the visibility flag."""
    if visible is True or (isinstance(visible, six.string_types) and visible[0].lower() == 'y'):
      visible = 'Yes'
    if visible is False:
      visible = 'No'
    return self.setOption('Visible', visible)

  def setExceptFileTypes(self, fileTypes):
    """Sets the expected file types."""
    if not isinstance(fileTypes, list):
      fileTypes = [fileTypes]
    self.__exceptFileTypes.update(fileTypes)
    self.setFileType([t for t in self.getFileTypeList() if t not in fileTypes])

  def getExceptFileTypes(self):
    return list(self.__exceptFileTypes)

  def getQueryDict(self):
    """Returns the bookkeeping dictionary."""
    return self.__bkQueryDict

  def getPath(self):
    """Returns the Bookkeeping path."""
    return self.__bkPath

  def makePath(self):
    """Builds a path from the dictionary."""
    bk = self.__bkQueryDict
    fileType = bk.get('FileType', '')
    if isinstance(fileType, list):
      fileType = ','.join(fileType)
    path = os.path.join('/',
                        bk.get('ConfigName', ''),
                        bk.get('ConfigVersion', ''),
                        bk.get('ConditionDescription', '.'),
                        bk.get('ProcessingPass', '.')[1:],
                        str(bk.get('EventType', '.')).replace('90000000', '.'),
                        fileType).replace('/./', '//')
    while True:
      if path.endswith('/'):
        path = path[:-1]
      else:
        return path

  def getFileTypeList(self):
    """Returns the file types."""
    fileTypes = self.__bkQueryDict.get('FileType', [])
    if not isinstance(fileTypes, list):
      fileTypes = [fileTypes]
    return fileTypes

  def getEventTypeList(self):
    """Returns the event types."""
    eventType = self.__bkQueryDict.get("EventType", [])
    if eventType:
      if not isinstance(eventType, list):
        eventType = [eventType]
    return eventType

  def getProcessingPass(self):
    """Returns the processing pass."""
    return self.__bkQueryDict.get('ProcessingPass', '')

  def getConditions(self):
    """Returns the Simulation/data taking conditions."""
    return self.__bkQueryDict.get('ConditionDescription', '')

  def getConfiguration(self):
    """Returns the configuration name and configuration version."""
    configName = self.__bkQueryDict.get('ConfigName', '')
    configVersion = self.__bkQueryDict.get('ConfigVersion', '')
    if not configName or not configVersion:
      return ''
    return os.path.join('/', configName, configVersion)

  def isVisible(self):
    """Returns True/False depending on the visibility flag."""
    return self.__bkQueryDict.get('Visible', 'All')

  def __fileType(self, fileType=None, returnList=False):
    """return the file types taking into account the expected file types."""
    gLogger.verbose("BKQuery.__fileType: %s, fileType: %s" % (self, fileType))
    if not fileType:
      return []
    self.__getAllBKFileTypes()
    if isinstance(fileType, list):
      fileTypes = fileType
    else:
      fileTypes = fileType.split(',')
    allRequested = None
    if fileTypes[0].lower() == "all":
      allRequested = True
      bkTypes = self.getBKFileTypes()
      gLogger.verbose('BKQuery.__fileType: bkTypes %s' % str(bkTypes))
      if bkTypes:
        fileTypes = list(set(bkTypes) - self.__exceptFileTypes)
      else:
        fileTypes = []
    expandedTypes = set()
    # print "Requested", fileTypes
    for ft in fileTypes:
      if ft.lower() == 'all.hist':
        allRequested = False
        expandedTypes.update([t for t in self.__exceptFileTypes.union(self.__bkFileTypes)
                              if t.endswith('HIST')])
      elif ft.lower().find("all.") == 0:
        ext = '.' + ft.split('.')[1]
        ft = []
        if allRequested is None:
          allRequested = True
        expandedTypes.update([t for t in set(self.getBKFileTypes()) - self.__exceptFileTypes
                              if t.endswith(ext)])
      else:
        expandedTypes.add(ft)
    # Remove __exceptFileTypes only if not explicitly required
    # print "Obtained", fileTypes, expandedTypes
    gLogger.verbose("BKQuery.__fileType: requested %s, expanded %s, except %s" % (allRequested,
                                                                                  expandedTypes,
                                                                                  self.__exceptFileTypes))
    if expandedTypes - self.__bkFileTypes and not self.__alreadyWarned:
      self.__alreadyWarned = True
      gLogger.always("**** Take care: some requested file types do not exist!!",
                     str(sorted(expandedTypes - self.__bkFileTypes)))
    if allRequested or not expandedTypes & self.__exceptFileTypes:
      expandedTypes -= self.__exceptFileTypes
    gLogger.verbose("BKQuery.__fileType: result %s" % sorted(expandedTypes))
    if len(expandedTypes) == 1 and not returnList:
      return list(expandedTypes)[0]
    return list(expandedTypes)

  def __getAllBKFileTypes(self):
    """Returns the file types from the bookkeeping database."""
    if not self.__bkFileTypes:
      self.__bkFileTypes = set([self.__fakeAllDST])
      warned = False
      while True:
        res = self.__bkClient.getAvailableFileTypes()
        if res['OK']:
          dbresult = res['Value']
          for record in dbresult['Records']:
            if record[0].endswith('HIST') or \
                    record[0].endswith('ETC') or \
                    record[0] == 'LOG' or \
                    record[0].endswith('ROOT'):
              self.__exceptFileTypes.add(record[0])
            self.__bkFileTypes.add(record[0])
          break
        if not warned:
          gLogger.always('Error getting BK file types, retrying', res['Message'])
          warned = True

  def __getBKFiles(self, bkQueryDict, retries=5):
    """Call BK getFiles() with some retries."""
    if not retries:
      retries = sys.maxsize
    errorLogged = False
    while retries:
      res = self.__bkClient.getFiles(bkQueryDict)
      if res['OK']:
        break
      retries -= 1
      if not errorLogged:
        errorLogged = True
        gLogger.warn("Error getting files from BK, retrying...", res['Message'])
    return res

  def getLFNsAndSize(self, getSize=True):
    """Returns the LFNs and their size for a given data set."""
    self.__getAllBKFileTypes()
    res = self.__getBKFiles(self.__bkQueryDict)
    lfns = []
    lfnSize = 0
    if not res['OK']:
      gLogger.error("Error from BK for %s:" % self.__bkQueryDict, res['Message'])
    else:
      lfns = set(res['Value'])
      exceptFiles = list(self.__exceptFileTypes)
      if exceptFiles and not self.__bkQueryDict.get('FileType'):
        res = self.__getBKFiles(BKQuery(self.__bkQueryDict).setOption('FileType', exceptFiles))
        if res['OK']:
          lfnsExcept = set(res['Value']) & lfns
        else:
          gLogger.error(
              "***** ERROR ***** Error in getting dataset from BK for %s files:" %
              exceptFiles, res['Message'])
          lfnsExcept = set()
        if lfnsExcept:
          gLogger.warn("***** WARNING ***** Found %d files in BK query that will be \
          excluded (file type in %s)!" % (len(lfnsExcept), str(exceptFiles)))
          gLogger.warn("                    If creating a transformation, set '--FileType ALL'")
          lfns = lfns - lfnsExcept
        else:
          exceptFiles = False
      if getSize:
        # Get size only if needed
        query = BKQuery(self.__bkQueryDict)
        query.setOption("FileSize", True)
        res = self.__getBKFiles(query.getQueryDict())
        if res['OK'] and isinstance(res['Value'], list) and res['Value'][0]:
          lfnSize = res['Value'][0]
        if exceptFiles and not self.__bkQueryDict.get('FileType'):
          res = self.__getBKFiles(query.setOption('FileType', exceptFiles))
          if res['OK'] and isinstance(res['Value'], list) and res['Value'][0]:
            lfnSize -= res['Value'][0]

        lfnSize /= 1000000000000.
      else:
        lfnSize = 0.
    return {'LFNs': list(lfns), 'LFNSize': lfnSize}

  def getLFNSize(self, visible=None):
    """Returns the size of a  given data set."""
    if visible is None:
      visible = self.isVisible()
    res = self.__getBKFiles(BKQuery(self.__bkQueryDict, visible=visible).setOption('FileSize', True))
    if res['OK'] and isinstance(res['Value'], list) and res['Value'][0]:
      lfnSize = res['Value'][0]
    else:
      lfnSize = 0
    return lfnSize

  def getNumberOfLFNs(self, visible=None):
    """Returns the number of LFNs correspond to a given data set."""
    if visible is None:
      visible = self.isVisible()
    if self.isVisible() != visible:
      query = BKQuery(self.__bkQueryDict, visible=visible)
    else:
      query = self
    fileTypes = query.getFileTypeList()
    nbFiles = 0
    size = 0

    for ft in fileTypes:
      if ft:
        res = self.__bkClient.getFilesSummary(query.setFileType(ft))
        # print query, res
        if res['OK']:
          res = res['Value']
          ind = res['ParameterNames'].index('NbofFiles')
          if res['Records'][0][ind]:
            nbFiles += res['Records'][0][ind]
            ind1 = res['ParameterNames'].index('FileSize')
            size += res['Records'][0][ind1]
            # print 'Visible',query.isVisible(),ft, 'Files:',
            # res['Records'][0][ind], 'Size:', res['Records'][0][ind1]
        else:
          gLogger.error("Error getting files summary", res['Message'])
          break
    return {'NumberOfLFNs': nbFiles, 'LFNSize': size}

  def getLFNs(self, printSEUsage=False, printOutput=True, visible=None):
    """returns a list of lfns.

    It prints statistics about the data sets if it is requested.
    """
    if visible is None:
      visible = self.isVisible()

    if self.isVisible() != visible:
      query = BKQuery(self.__bkQueryDict, visible=visible)
    else:
      query = self

    # Loop for each production or each event type rather than make a single query
    loopItem = None
    prods = self.__bkQueryDict.get('Production')
    eventTypes = self.__bkQueryDict.get('EventType')
    if prods and isinstance(prods, list):
      loopItem = 'Production'
      loopList = prods
    elif eventTypes and isinstance(eventTypes, list):
      loopItem = 'EventType'
      loopList = eventTypes
    if loopItem:
      # It's faster to loop on a list of prods or event types than query the BK with a list as argument
      lfns = []
      lfnSize = 0
      if query == self:
        query = BKQuery(self.__bkQueryDict, visible=visible)
      for item in loopList:
        query.setOption(loopItem, item)
        lfnsAndSize = query.getLFNsAndSize(getSize=printOutput)
        lfns += lfnsAndSize['LFNs']
        lfnSize += lfnsAndSize['LFNSize']
    else:
      lfnsAndSize = query.getLFNsAndSize(getSize=printOutput)
      lfns = lfnsAndSize['LFNs']
      lfnSize = lfnsAndSize['LFNSize']

    if not lfns:
      gLogger.verbose("No files found for BK query %s" % str(self.__bkQueryDict))
    else:
      lfns.sort()

      # Only for printing
      if printOutput:
        gLogger.notice("\n%d files (%.1f TB) in directories:" % (len(lfns), lfnSize))
        dirs = {}
        for lfn in lfns:
          directory = os.path.join(os.path.dirname(lfn), '')
          dirs[directory] = dirs.setdefault(directory, 0) + 1
        for directory in sorted(dirs):
          gLogger.notice("%s %s files" % (directory, dirs[directory]))
        if printSEUsage:
          rpc = StorageUsageClient()
          totalUsage = {}
          totalSize = 0
          for directory in dirs:
            res = rpc.getStorageSummary(directory, '', '', [])
            if res['OK']:
              for se in [se for se in res['Value'] if not se.endswith("-ARCHIVE")]:
                totalUsage[se] = totalUsage.setdefault(se, 0) + res['Value'][se]['Size']
                totalSize += res['Value'][se]['Size']
          ses = sorted(totalUsage)
          totalUsage['Total'] = totalSize
          ses.append('Total')
          gLogger.notice("\n%s %s" % ("SE".ljust(20), "Size (TB)"))
          for se in ses:
            gLogger.notice("%s %s" % (se.ljust(20), ('%.1f' % (totalUsage[se] / 1000000000000.))))
    return lfns

  def getDirs(self, printOutput=False, visible=None):
    """Returns the directories."""
    if visible is None:
      visible = self.isVisible()
    lfns = self.getLFNs(printSEUsage=True, printOutput=printOutput, visible=visible)
    dirs = set()
    for lfn in lfns:
      dirs.add(os.path.dirname(lfn))
    return sorted(dirs)

  @staticmethod
  def __getProdStatus(prod):
    """Returns the status of a given transformation."""
    res = TransformationClient().getTransformation(prod, extraParams=False)
    if not res['OK']:
      gLogger.error("Could not get information", "on production %d" % prod)
      return None
    return res['Value']['Status']

  def getBKRuns(self):
    """It returns a list of runs from the bookkeeping."""
    if self.getProcessingPass().replace('/', '') == 'Real Data':
      return self.getBKProductions()

  def getBKProductions(self, visible=None):
    """It returns a list of productions."""
    if visible is None:
      visible = self.isVisible()
    prodList = self.__bkQueryDict.get('Production')
    if prodList:
      if not isinstance(prodList, list):
        prodList = [prodList]
      return sorted(prodList)
    if not self.getProcessingPass():
      gLogger.fatal('Impossible to get a list of productions without the Processing Pass')
      return []
    eventTypes = self.__bkQueryDict.get('EventType')
    if not isinstance(eventTypes, list):
      eventTypes = [eventTypes]
    fullList = set()
    for eventType in eventTypes:
      bkQ = BKQuery(self.__bkQueryDict)
      bkQ.setVisible(visible)
      bkDict = bkQ.setEventType(eventType)
      # gLogger.notice( 'Get productions for BK query', str( bkDict ) )
      res = self.__bkClient.getProductions(bkDict)
      if not res['OK']:
        gLogger.error('Error getting productions from BK', res['Message'])
        return []
      if self.getProcessingPass().replace('/', '') != 'Real Data':
        fileTypes = self.getFileTypeList()
        prodList = set(prod for prods in res['Value']['Records'] for prod in prods
                       if self.__getProdStatus(prod) != 'Deleted')
        # print '\n', self.__bkQueryDict, res['Value']['Records'], '\nVisible:', visible, prodList
        pList = set()
        if fileTypes:
          transClient = TransformationClient()
          for prod in prodList:
            res = transClient.getBookkeepingQuery(prod)
            if res['OK'] and res['Value']['FileType'] in fileTypes:
              pList.add(prod)
        if not pList:
          pList = prodList
      else:
        runList = sorted([-run for r in res['Value']['Records'] for run in r])
        startRun = int(self.__bkQueryDict.get('StartRun', 0))
        endRun = int(self.__bkQueryDict.get('EndRun', sys.maxsize))
        pList = set(run for run in runList if run >= startRun and run <= endRun)
      fullList.update(pList)
    return sorted(fullList)

  def getBKConditions(self):
    """It returns the data taking / simulation conditions."""
    conditions = self.__bkQueryDict.get('ConditionDescription')
    if conditions:
      if not isinstance(conditions, list):
        conditions = [conditions]
      return conditions
    result = self.__bkClient.getConditions(self.__bkQueryDict)
    if result['OK']:
      resList = result['Value']
    else:
      return []
    conditions = []
    for res in resList:
      ind = res['ParameterNames'].index('Description')
      if res['Records']:
        conditions += [par[ind] for par in res['Records']]
        break
    return sorted(conditions)

  def getBKEventTypes(self):
    """It returns the event types."""
    eventType = self.getEventTypeList()
    if eventType:
      return eventType
    res = self.__bkClient.getEventTypes(self.__bkQueryDict)['Value']
    ind = res['ParameterNames'].index('EventType')
    eventTypes = sorted(rec[ind] for rec in res['Records'])
    return eventTypes

  def getBKFileTypes(self, bkDict=None):
    """It returns the file types."""
    fileTypes = self.getFileTypeList()
    # print "Call getBKFileTypes:", self, fileTypes
    if not fileTypes:
      if not bkDict:
        bkDict = self.__bkQueryDict.copy()
      else:
        bkDict = bkDict.copy()
      bkDict.setdefault('Visible', 'All')
      bkDict['ReplicaFlag'] = 'All'
      bkDict.pop('RunNumber', None)
      fileTypes = []
      eventTypes = bkDict.get('EventType')
      if isinstance(eventTypes, list):
        for et in eventTypes:
          bkDict['EventType'] = et
          fileTypes += self.getBKFileTypes(bkDict)
      else:
        res = self.__bkClient.getFileTypes(bkDict)
        if res['OK']:
          res = res['Value']
          ind = res['ParameterNames'].index('FileTypes')
          fileTypes = [rec[ind] for rec in res['Records'] if rec[ind] not in self.__exceptFileTypes]
    if 'ALL.DST' in fileTypes:
      fileTypes.remove('ALL.DST')
      fileTypes.append(self.__fakeAllDST)
    # print 'FileTypes1', fileTypes
    fileTypes = self.__fileType(fileTypes, returnList=True)
    # print 'FileTypes2', fileTypes
    if self.__fakeAllDST in fileTypes:
      fileTypes.remove(self.__fakeAllDST)
      fileTypes.append('ALL.DST')
    # print 'FileTypes3', fileTypes
    return fileTypes

  def getBKProcessingPasses(self, queryDict=None, depth=None):
    """It returns the processing pass."""
    if depth is None:
      depth = sys.maxsize
    processingPasses = {}
    if not queryDict:
      queryDict = self.__bkQueryDict.copy()
    initialPP = queryDict.get('ProcessingPass', '/')
    # print "Start", initialPP, queryDict
    res = self.__bkClient.getProcessingPass(queryDict, initialPP)
    if not res['OK']:
      if 'Empty Directory' not in res['Message']:
        gLogger.error("ERROR getting processing passes for %s" % queryDict, res['Message'])
      return {}
    ppRecords = res['Value'][0]
    if 'Name' in ppRecords['ParameterNames']:
      ind = ppRecords['ParameterNames'].index('Name')
      passes = sorted([os.path.join(initialPP, rec[ind]) for rec in ppRecords['Records']])
    else:
      passes = []
    evtRecords = res['Value'][1]
    if 'EventType' in evtRecords['ParameterNames']:
      ind = evtRecords['ParameterNames'].index('EventType')
      eventTypes = [str(rec[ind]) for rec in evtRecords['Records']]
    else:
      eventTypes = []

    if passes and depth:
      depth -= 1
      nextProcessingPasses = {}
      for pName in passes:
        processingPasses[pName] = []
        queryDict['ProcessingPass'] = pName
        nextProcessingPasses.update(self.getBKProcessingPasses(queryDict, depth=depth))
      processingPasses.update(nextProcessingPasses)
    if eventTypes:
      processingPasses[initialPP] = eventTypes
    for pName in ('/Real Data', '/'):
      if pName in processingPasses:
        processingPasses.pop(pName)
    # print "End", initialPP, [(key, processingPasses[key]) for key in sorted(list(processingPasses))]
    return processingPasses



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Utility for invoking running LHCb applications."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys
import os
import shlex

from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.Subprocess import systemCall


class LbRunError(RuntimeError):
  """Exception for lb-run errors."""
  pass


class LHCbApplicationError(RuntimeError):
  """Exception for application errors."""
  pass


class LHCbDIRACError(RuntimeError):
  """Exception for application errors."""
  pass


class RunApplication(object):
  """Encapsulate logic for running an LHCb application."""

  def __init__(self):
    """ c'tor - holds common variables
    """
    # Standard LHCb scripts
    self.runApp = 'lb-run'
    self.lbrunCommand = ''  # Command that will be constructed at run time

    # What to run
    self.applicationName = ''  # e.g. Gauss
    self.applicationVersion = ''  # e.g v42r1

    # Define the environment
    self.extraPackages = []
    self.systemConfig = 'ANY'
    self.runTimeProject = ''
    self.runTimeProjectVersion = ''
    self.site = ''

    # What to run and how
    self.command = 'gaudirun.py'
    self.commandOptions = []
    self.prodConf = False
    self.prodConfFileName = 'prodConf.py'
    self.step_number = 0
    self.extraOptionsLine = ''
    self.numberOfProcessors = 1

    self.applicationLog = 'applicationLog.txt'
    self.stdError = 'applicationError.txt'

    # Utilities
    self.log = gLogger.getSubLogger("RunApplication")
    self.opsH = Operations()

    # Prmon
    self.prmonPath = '/cvmfs/lhcb.cern.ch/lib/experimental/prmon/bin/prmon'
    self.usePrmon = False

  def run(self):
    """Invokes lb-run (what you call after having setup the object)"""
    self.log.info("Executing application %s %s for binary tag configuration '%s'" % (self.applicationName,
                                                                                     self.applicationVersion,
                                                                                     self.systemConfig))

    lbRunOptions = self.opsH.getValue('GaudiExecution/lbRunOptions', '--siteroot=/cvmfs/lhcb.cern.ch/lib/')

    extraPackagesString, runtimeProjectString, externalsString = self._lbRunCommandOptions()

    # if a binary tag is provided, then only that should be called.
    # This should be safeguarded with the following method.
    # if not, we call "-c best"
    if self.systemConfig.lower() == 'any':
      self.systemConfig = 'best'
    configString = "-c %s" % self.systemConfig

    # App
    app = self.applicationName
    if self.applicationVersion:
      app += '/' + self.applicationVersion

    # Command
    if self.command == 'gaudirun.py':
      command = self._gaudirunCommand()
    else:
      command = self.command

    self.lbrunCommand = ' '.join([self.runApp, lbRunOptions,
                                  configString, extraPackagesString,
                                  runtimeProjectString, externalsString,
                                  app])

    finalCommand = ' '.join([self.lbrunCommand, command])

    # then run it!
    runResult = self._runApp(finalCommand, env=self._getEnv())
    if not runResult['OK']:
      self.log.error("Problem executing lb-run: %s" % runResult['Message'])
      self.log.error("Environment: %s" % os.environ)
      raise LHCbDIRACError("Can not start %s %s" % (self.applicationName, self.applicationVersion))

    if runResult['Value'][0]:  # if exit status != 0
      self.log.error("lb-run or its application exited with status %d" % runResult['Value'][0])
      self.log.error(runResult['Value'][2])

      # this is an lb-run specific error
      if runResult['Value'][0] & 0x40 and not runResult['Value'][0] & 0x80:
        self.log.error("Status %d is an lb-run specific error" % runResult['Value'][0])
        self.log.error("Trying running the same lb-run command with --debug option, it should fail immediately anyway")
        lbrunCommandWithDebug = ' '.join([self.runApp, '--debug', lbRunOptions,
                                          configString, extraPackagesString,
                                          runtimeProjectString, externalsString,
                                          app])
        debugCommand = ' '.join([lbrunCommandWithDebug, command])
        self._runApp(debugCommand)
        raise LbRunError("Problem setting the environment: lb-run exited with status %d" % runResult['Value'][0])

      self.log.error("Status %d is an application (%s %s) error" % (runResult['Value'][0],
                                                                    self.applicationName,
                                                                    self.applicationVersion))
      raise LHCbApplicationError("%s %s exited with status %d" % (self.applicationName,
                                                                  self.applicationVersion,
                                                                  runResult['Value'][0]))

    self.log.info("%s execution completed successfully" % self.applicationName)

    return runResult

  def _lbRunCommandOptions(self):
    """Return lb-run command options."""

    # extra packages (for setup phase) (added using '--use')
    extraPackagesString = ''
    for ep in self.extraPackages:
      if len(ep) == 1:
        epName = ep[0]
        epVer = ''
      elif len(ep) == 2:
        epName, epVer = ep

      if epName.lower() == 'prodconf':
        self.prodConf = True
      if epVer:
        extraPackagesString += ' --use="%s %s" ' % (epName, epVer)
      else:
        extraPackagesString += ' --use="%s"' % epName

    # run time project
    runtimeProjectString = ''
    if self.runTimeProject:
      self.log.verbose('Requested run time project: %s' % (self.runTimeProject))
      runtimeProjectString = ' --runtime-project %s/%s' % (self.runTimeProject, self.runTimeProjectVersion)

    externalsString = ''

    externals = []
    if self.site:
      externals = self.opsH.getValue('ExternalsPolicy/%s' % (self.site), [])
      if externals:
        self.log.info('Found externals policy for %s = %s' % (self.site, externals))
        for external in externals:
          externalsString += ' --ext=%s' % external

    if not externalsString:
      externals = self.opsH.getValue('ExternalsPolicy/Default', [])
      if externals:
        self.log.info('Using default externals policy for %s = %s' % (self.site, externals))
        for external in externals:
          externalsString += ' --ext=%s' % external

    return extraPackagesString, runtimeProjectString, externalsString

  def _gaudirunCommand(self):
    """construct a gaudirun command."""
    command = self.opsH.getValue('/GaudiExecution/gaudirunFlags', 'gaudirun.py')

    # multicore?
    if self.numberOfProcessors > 1:
      command += ' --ncpus %d ' % int(self.numberOfProcessors)

    if self.commandOptions:
      command += ' '
      command += ' '.join(self.commandOptions)

    if self.prodConf:
      command += ' '
      command += self.prodConfFileName

    if self.extraOptionsLine:
      command += ' '
      with open('gaudi_extra_options.py', 'w') as fopen:
        fopen.write(self.extraOptionsLine)
      command += 'gaudi_extra_options.py'

    return command

  def _runApp(self, command, env=None):
    """Safe system call of a command.

    :param command basestring: the command to run
    :param env dict: environment where to run -- maybe the LHCb environment from LbLogin
    """
    if self.applicationName == 'Gauss' and self.usePrmon:
      command = self.prmonPath + ' --json-summary ./prmon_Gauss.json -- ' + command

    print('Command called: \n%s' % command)  # Really printing here as we want to see and maybe cut/paste

    return systemCall(timeout=0,
                      cmdSeq=shlex.split(command),
                      callbackFunction=self.__redirectLogOutput,
                      env=env)

  def _getEnv(self):
    """Get a dictionary containing the environment that should be used for the job
    """
    # Start from the current environment
    env = os.environ.copy()
    # Versions of Brunel used for 2018 data use XGBoost which uses OpenMP to
    # provide parallelism and automatically spawns one thread for each CPU.
    # Use OMP_NUM_THREADS to force it to only use one thread
    env["OMP_NUM_THREADS"] = str(self.numberOfProcessors)
    return env

  def __redirectLogOutput(self, fd, message):
    """Callback function for the Subprocess calls (manages log files)

    Args:
        fd (int): stdin/stderr file descriptor
        message (str): line to log
    """
    sys.stdout.flush()
    if message:
      if 'INFO Evt' in message or 'Reading Event record' in message:  # These ones will appear in the std.out log too
        print(message)
      if self.applicationLog:
        with open(self.applicationLog, 'a') as log:
          log.write(message + '\n')
          log.flush()
      else:
        self.log.error("Application Log file not defined")
      if fd == 1:
        if self.stdError:
          with open(self.stdError, 'a') as error:
            error.write(message + '\n')
            error.flush()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get the GUID of a (set of) ROOT file The file can be either local, an LFN or
an xrootd URL (root:...)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] file1 [file2 ...]' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)
  files = []
  for oFile in Script.getPositionalArgs():
    files += oFile.split(',')

  import DIRAC
  from DIRAC.Interfaces.API.Dirac import Dirac
  from LHCbDIRAC.Core.Utilities.File import getRootFileGUIDs
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import printDMResult

  if not files:
    Script.showHelp(exitCode=1)
  existFiles = {}
  nonExisting = []
  dirac = Dirac()

  for localFile in files:
    if os.path.exists(localFile):
      existFiles[os.path.realpath(localFile)] = localFile
    elif localFile.startswith('/lhcb'):
      res = dirac.getReplicas(localFile, active=True, preferDisk=True)
      if res['OK'] and localFile in res['Value']['Successful']:
        ses = list(res['Value']['Successful'][localFile])
        for se in ses:
          res = dirac.getAccessURL(localFile, se, protocol=['root', 'xroot'])
          if res['OK'] and localFile in res['Value']['Successful']:
            existFiles[res['Value']['Successful'][localFile]] = "%s @ %s" % (localFile, se)
      else:
        nonExisting.append(localFile)
    elif localFile.startswith('root:'):
      existFiles[localFile] = localFile
    else:
      nonExisting.append(localFile)

  fileGUIDs = getRootFileGUIDs(list(existFiles))
  for status in ('Successful', 'Failed'):
    for file in list(fileGUIDs.get('Value', {}).get(status, {})):
      fileGUIDs['Value'][status][existFiles.get(file, file)] = fileGUIDs['Value'][status].pop(file)
  if nonExisting:
    fileGUIDs['Value']['Failed'].update(dict.fromkeys(nonExisting, 'Non existing file'))
  printDMResult(fileGUIDs)

  DIRAC.exit(0)


if __name__ == "__main__":
  main()
